#  driver.py
#  HammerDriver and related code.
#
#  See LICENSE for licence details.

from functools import reduce, partial
from typing import NamedTuple, List, Optional, Tuple, Dict, Set, Any
from pathlib import Path

import datetime
import importlib
import os

from hammer.utils import *

import hammer.config as hammer_config
import hammer.tech as hammer_tech
from hammer.tech import MacroSize
from .hammer_tool import HammerTool
from .hooks import HammerToolHookAction
from .hammer_vlsi_impl import HammerVLSISettings, HammerPlaceAndRouteTool, HammerSynthesisTool, \
    HammerSignoffTool, HammerDRCTool, HammerLVSTool, HammerSRAMGeneratorTool, HammerPCBDeliverableTool, HammerSimTool, HammerPowerTool, HammerFormalTool, HammerTimingTool, \
    HierarchicalMode, load_tool, PlacementConstraint, SRAMParameters, ILMStruct, FlowLevel
from hammer.logging import HammerVLSIFileLogger, HammerVLSILogging, HammerVLSILoggingContext
from .submit_command import HammerSubmitCommand

__all__ = ['HammerDriverOptions', 'HammerDriver']

# Options for invoking the driver.
HammerDriverOptions = NamedTuple('HammerDriverOptions', [
    # List of environment config files in .json
    ('environment_configs', List[str]),
    # List of project config files in .json or .yml
    ('project_configs', List[str]),
    # Log file location.
    ('log_file', str),
    # Folder for storing runtime files / CAD junk.
    ('obj_dir', str)
])


class HammerDriver:
    @staticmethod
    def get_default_driver_options() -> HammerDriverOptions:
        """Get default driver options."""
        return HammerDriverOptions(
            environment_configs=[],
            project_configs=[],
            log_file=datetime.datetime.now().strftime("hammer-vlsi-%Y%m%d-%H%M%S.log"),
            obj_dir=str((Path.cwd() / "obj_dir").resolve())
        )

    def __init__(self, options: HammerDriverOptions, extra_project_config: dict = {}) -> None:
        """
        Create a hammer-vlsi driver, which is a higher level convenience function
        for quickly using hammer-vlsi. It imports and uses the hammer-vlsi blocks.

        Set up logging, databases, context, etc.

        :param options: Driver options.
        :param extra_project_config: An extra flattened config for the project. Optional.
        """

        # Create global logging context.
        file_logger = HammerVLSIFileLogger(options.log_file)
        HammerVLSILogging.add_callback(file_logger.callback)
        self.log = HammerVLSILogging.context()  # type: HammerVLSILoggingContext

        # Create a new hammer database.
        self.database = hammer_config.HammerDatabase()  # type: hammer_config.HammerDatabase

        self.log.info("Loading hammer-vlsi libraries and reading settings")

        # Store the run dir (this should already be canonicalized by the CLI driver).
        self.obj_dir = options.obj_dir  # type: str

        # Also store the options
        self.options = options

        # Load builtins and core into the database.
        HammerVLSISettings.load_builtins_and_core(self.database)

        # Read in the environment config for paths to CAD tools, etc.
        for config in options.environment_configs:
            if not os.path.exists(config):
                self.log.error("Environment config %s does not exist!" % (config))
            config_str = Path(config).read_text()
            is_yaml = config.endswith(".yml")
            self.database.update_environment([hammer_config.load_config_from_string(config_str, is_yaml, str(Path(config).resolve().parent))])

        # Read in the project config to find the syn, par, and tech.
        project_configs: List[dict] = []
        for config in options.project_configs:
            if not os.path.exists(config):
                self.log.error("Project config %s does not exist!" % (config))
            config_str = Path(config).read_text()
            is_yaml = config.endswith(".yml")
            project_configs.append(hammer_config.load_config_from_string(config_str, is_yaml, str(Path(config).resolve().parent)))
        project_configs.append(extra_project_config)
        self.project_configs = []  # type: List[dict]
        self.update_project_configs(project_configs)

        # Get the technology and load technology settings.
        self.tech = None  # type: Optional[hammer_tech.HammerTechnology]
        self.load_technology()

        # Keep track of what the synthesis and par configs are since
        # update_tools() just takes a whole list.
        self.tool_configs = {}  # type: Dict[str, List[dict]]
        self.tool_config_types = {}  # type: Dict[str, List[dict]]

        # Initialize tool fields.
        self.syn_tool = None  # type: Optional[HammerSynthesisTool]
        self.par_tool = None  # type: Optional[HammerPlaceAndRouteTool]
        self.drc_tool = None  # type: Optional[HammerDRCTool]
        self.lvs_tool = None  # type: Optional[HammerLVSTool]
        self.sram_generator_tool = None  # type: Optional[HammerSRAMGeneratorTool]
        self.sim_tool = None  # type: Optional[HammerSimTool]
        self.power_tool = None # type: Optional[HammerPowerTool]
        self.formal_tool = None # type: Optional[HammerFormalTool]
        self.timing_tool = None # type: Optional[HammerTimingTool]

        # Initialize tool hooks. Used to specify resume/pause hooks after custom hooks have been registered.
        self.post_custom_syn_tool_hooks = []  # type: List[HammerToolHookAction]
        self.post_custom_par_tool_hooks = []  # type: List[HammerToolHookAction]
        self.post_custom_drc_tool_hooks = []  # type: List[HammerToolHookAction]
        self.post_custom_lvs_tool_hooks = []  # type: List[HammerToolHookAction]
        self.post_custom_sram_generator_tool_hooks = []  # type: List[HammerToolHookAction]
        self.post_custom_sim_tool_hooks = []  # type: List[HammerToolHookAction]
        self.post_custom_power_tool_hooks = [] # type: List[HammerToolHookAction]
        self.post_custom_formal_tool_hooks = [] # type: List[HammerToolHookAction]
        self.post_custom_timing_tool_hooks = [] # type: List[HammerToolHookAction]
        self.post_custom_pcb_tool_hooks = []  # type: List[HammerToolHookAction]

        self._dump_history = False

    @property
    def project_config(self) -> dict:
        return hammer_config.combine_configs(self.project_configs)

    def update_project_configs(self, project_configs: List[dict]) -> None:
        """
        Update the project configs in the driver and database.
        """
        self.project_configs = project_configs
        self.database.update_project(self.project_configs)

    def load_technology(self, cache_dir: str = "") -> None:
        tech_module: str = self.database.get_setting("vlsi.core.technology")
        tech_name = tech_module.split('.')[-1]

        if cache_dir == "":
            cache_dir = os.path.join(self.obj_dir, "tech-%s-cache" % tech_name)

        self.log.info("Loading technology '{0}'".format(tech_module))
        tech_opt = hammer_tech.HammerTechnology.load_from_module(tech_module)
        if tech_opt is None:
            self.log.fatal("Technology {0} not found or missing .tech.[json/yml]!".format(tech_module))
            return
        else:
            tech: hammer_tech.HammerTechnology = tech_opt
        # Update database as soon as possible since e.g. extract_technology_files could use those settings
        self.database.update_technology(*tech.get_config())
        tech.logger = self.log.context("tech")
        tech.set_database(self.database)
        tech.cache_dir = cache_dir
        tech.extract_technology_files()

        self.tech = tech

    def update_tool_configs(self) -> None:
        """
        Calls self.database.update_tools with self.tool_configs as a list.
        """
        tools = reduce(lambda a, b: a + b, list(self.tool_configs.values()))
        types = reduce(lambda a, b: a + b, list(self.tool_config_types.values()))
        self.database.update_tools(tools, types)

    def instantiate_tool_from_config(self, tool_type: str,
                                     required_type: Optional[type] = None) -> Optional[Tuple[HammerTool, str]]:
        """
        Create a new instance of the given tool using information from the config.
        :param tool_type: Tool type. e.g. if "par", then this will look in vlsi.core.par_tool
        :param required_type: (optional) Check that the instantiated tool is the given type.
        :return: Tuple of (tool instance, tool name) or
                 None if an error occurred.
        """
        # Find the tool and read in their configs.
        tool_module = self.database.get_setting("vlsi.core.{tool_type}_tool".format(tool_type=tool_type))
        tool_get = load_tool(tool_module)
        if required_type is not None:
            if not isinstance(tool_get, required_type):
                self.log.error("{tool_type} tool's type is incorrect: got {got}".format(tool_type=tool_type,
                                                                                        got=str(type(tool_get))))
                return None
        tool_name = tool_module.split('.')[-1]
        return tool_get, tool_name

    def set_up_synthesis_tool(self, syn_tool: HammerSynthesisTool,
                              name: str, run_dir: str = "") -> bool:
        """
        Set up and store the given synthesis tool instance for use in this
        driver.
        :param syn_tool: Tool instance.
        :param name: Short name (e.g. "yosys") of the tool instance. Typically
                     obtained from the database.
        :param run_dir: Directory to use for the tool run_dir. Defaults to the
                        run_dir passed in the HammerDriver constructor.
        :return: True if setup was successful.
        """
        if self.tech is None:
            self.log.error("Must load technology before loading synthesis tool")
            return False

        if run_dir == "":
            run_dir = os.path.join(self.obj_dir, "syn-rundir")
        # TODO: generate this automatically
        syn_tool.name = name
        syn_tool.logger = self.log.context("synthesis")
        syn_tool.technology = self.tech
        syn_tool.set_database(self.database)
        syn_tool.run_dir = run_dir
        syn_tool.hierarchical_mode = HierarchicalMode.from_str(
            self.database.get_setting("vlsi.inputs.hierarchical.mode"))
        syn_tool.input_files = self.database.get_setting("synthesis.inputs.input_files")
        syn_tool.top_module = self.database.get_setting("synthesis.inputs.top_module", nullvalue="")
        syn_tool.submit_command = HammerSubmitCommand.get("synthesis", self.database)
        syn_tool.output_all_regs = ""
        syn_tool.output_seq_cells = ""

        # TODO: automate this based on the definitions
        missing_inputs = False
        if syn_tool.top_module == "":
            self.log.error("Top module not specified for synthesis")
            missing_inputs = True
        if len(syn_tool.input_files) == 0:
            self.log.error("No input files specified for synthesis")
            missing_inputs = True
        if missing_inputs:
            return False

        self.syn_tool = syn_tool
        self.tool_configs["synthesis"], self.tool_config_types["synthesis"] = syn_tool.get_config()
        self.update_tool_configs()
        return True

    def load_synthesis_tool(self, run_dir: str = "") -> bool:
        """
        Load the synthesis tool based on the given database.

        :param run_dir: Directory to use for the tool run_dir. Defaults to the run_dir passed in the HammerDriver
                        constructor.
        :return: True if synthesis tool loading was successful, False otherwise.
        """
        config_result = self.instantiate_tool_from_config("synthesis", HammerSynthesisTool)
        if config_result is None:
            return False
        else:
            (syn_tool, name) = config_result
            assert isinstance(syn_tool, HammerSynthesisTool)
            return self.set_up_synthesis_tool(syn_tool, name, run_dir)

    def set_up_par_tool(self, par_tool: HammerPlaceAndRouteTool,
                        name: str, run_dir: str = "") -> bool:
        """
        Set up and store the given place-and-route tool instance for use in this
        driver.
        :param par_tool: Tool instance.
        :param name: Short name (e.g. "yosys") of the tool instance. Typically
                     obtained from the database.
        :param run_dir: Directory to use for the tool run_dir. Defaults to the
                        run_dir passed in the HammerDriver constructor.
        :return: True if setup was successful.
        """
        if self.tech is None:
            self.log.error("Must load technology before loading par tool")
            return False

        if run_dir == "":
            run_dir = os.path.join(self.obj_dir, "par-rundir")

        par_tool.name = name
        par_tool.logger = self.log.context("par")
        par_tool.technology = self.tech
        par_tool.set_database(self.database)
        par_tool.run_dir = run_dir
        par_tool.hierarchical_mode = HierarchicalMode.from_str(
            self.database.get_setting("vlsi.inputs.hierarchical.mode"))
        par_tool.submit_command = HammerSubmitCommand.get("par", self.database)

        missing_inputs = False

        # TODO: automate this based on the definitions
        par_tool.input_files = list(self.database.get_setting("par.inputs.input_files"))
        par_tool.top_module = self.database.get_setting("par.inputs.top_module", nullvalue="")
        par_tool.post_synth_sdc = self.database.get_setting("par.inputs.post_synth_sdc", nullvalue="")
        par_tool.output_all_regs = ""
        par_tool.output_seq_cells = ""


        if len(par_tool.input_files) == 0:
            self.log.error("No input files specified for par")
            missing_inputs = True
        if par_tool.top_module == "":
            self.log.error("No top module specified for par")
            missing_inputs = True
        if missing_inputs:
            return False

        self.par_tool = par_tool
        self.tool_configs["par"], self.tool_config_types["par"] = par_tool.get_config()
        self.update_tool_configs()
        return True

    def load_par_tool(self, run_dir: str = "") -> bool:
        """
        Load the place and route tool based on the given database.

        :param run_dir: Directory to use for the tool run_dir. Defaults to the run_dir passed in the HammerDriver
                        constructor.
        :return: True if successful, false otherwise
        """
        config_result = self.instantiate_tool_from_config("par", HammerPlaceAndRouteTool)
        if config_result is None:
            return False
        else:
            (par_tool, name) = config_result
            assert isinstance(par_tool, HammerPlaceAndRouteTool)
            return self.set_up_par_tool(par_tool, name, run_dir)

    def set_up_drc_tool(self, drc_tool: HammerDRCTool,
                        name: str, run_dir: str = "") -> bool:
        """
        Set up and store the given DRC tool instance for use in this
        driver.
        :param drc_tool: Tool instance.
        :param name: Short name (e.g. "calibre") of the tool instance. Typically
                     obtained from the database.
        :param run_dir: Directory to use for the tool run_dir. Defaults to the
                        run_dir passed in the HammerDriver constructor.
        :return: True if setup was successful.
        """
        if self.tech is None:
            self.log.error("Must load technology before loading DRC tool")
            return False

        if run_dir == "":
            run_dir = os.path.join(self.obj_dir, "drc-rundir")

        drc_tool.name = name
        drc_tool.logger = self.log.context("drc")
        drc_tool.technology = self.tech
        drc_tool.set_database(self.database)
        drc_tool.hierarchical_mode = HierarchicalMode.from_str(
            self.database.get_setting("vlsi.inputs.hierarchical.mode"))
        drc_tool.submit_command = HammerSubmitCommand.get("drc", self.database)
        drc_tool.run_dir = run_dir
        # TODO hierarchical

        drc_tool.top_module = self.database.get_setting("drc.inputs.top_module", nullvalue="")
        drc_tool.layout_file = self.database.get_setting("drc.inputs.layout_file", nullvalue="")
        missing_inputs = False
        if drc_tool.top_module == "":
            self.log.error("Top module not specified for DRC")
            missing_inputs = True
        if drc_tool.layout_file is None:
            self.log.error("No layout file specified for DRC")
            missing_inputs = True
        if missing_inputs:
            return False

        self.drc_tool = drc_tool
        self.tool_configs["drc"], self.tool_config_types["drc"] = drc_tool.get_config()
        self.update_tool_configs()
        return True

    def load_drc_tool(self, run_dir: str = "") -> bool:
        """
        Loads a DRC tool on a given database

        :param run_dir: Directory to use for the tool run_dir. Defaults to the run_dir passed in the HammerDriver
                        constructor.
        :return: True if DRC tool loading was successful, False otherwise.
        """
        config_result = self.instantiate_tool_from_config("drc", HammerDRCTool)
        if config_result is None:
            return False
        else:
            (drc_tool, name) = config_result
            assert isinstance(drc_tool, HammerDRCTool)
            return self.set_up_drc_tool(drc_tool, name, run_dir)

    def set_up_lvs_tool(self, lvs_tool: HammerLVSTool,
                        name: str, run_dir: str = "") -> bool:
        """
        Set up and store the given LVS tool instance for use in this
        driver.
        :param lvs_tool: Tool instance.
        :param name: Short name (e.g. "calibre") of the tool instance. Typically
                     obtained from the database.
        :param run_dir: Directory to use for the tool run_dir. Defaults to the
                        run_dir passed in the HammerDriver constructor.
        :return: True if setup was successful.
        """
        if self.tech is None:
            self.log.error("Must load technology before loading LVS tool")
            return False

        if run_dir == "":
            run_dir = os.path.join(self.obj_dir, "lvs-rundir")

        lvs_tool.name = name
        lvs_tool.logger = self.log.context("lvs")
        lvs_tool.technology = self.tech
        lvs_tool.set_database(self.database)
        lvs_tool.hierarchical_mode = HierarchicalMode.from_str(
            self.database.get_setting("vlsi.inputs.hierarchical.mode"))
        lvs_tool.submit_command = HammerSubmitCommand.get("lvs", self.database)
        lvs_tool.run_dir = run_dir

        lvs_tool.schematic_files = self.database.get_setting("lvs.inputs.schematic_files", nullvalue=[])
        lvs_tool.layout_file = self.database.get_setting("lvs.inputs.layout_file", nullvalue="")
        lvs_tool.top_module = self.database.get_setting("lvs.inputs.top_module", nullvalue="")
        lvs_tool.hcells_list = self.database.get_setting("lvs.inputs.hcells_list", nullvalue=[])
        lvs_tool.ilms = list(map(lambda x: ILMStruct.from_setting(x), self.database.get_setting("lvs.inputs.ilms", nullvalue=[])))
        missing_inputs = False
        if lvs_tool.top_module == "":
            self.log.error("Top module not specified for LVS")
            missing_inputs = True
        if lvs_tool.layout_file is None:
            self.log.error("No layout file specified for LVS")
            missing_inputs = True
        if len(lvs_tool.schematic_files) == 0:
            self.log.error("No schematic files specified for LVS")
            missing_inputs = True
        if missing_inputs:
            return False

        self.lvs_tool = lvs_tool
        self.tool_configs["lvs"], self.tool_config_types["lvs"] = lvs_tool.get_config()
        self.update_tool_configs()
        return True

    def load_lvs_tool(self, run_dir: str = "") -> bool:
        """
        Loads an LVS tool on a given database

        :param run_dir: Directory to use for the tool run_dir. Defaults to the run_dir passed in the HammerDriver
                        constructor.
        :return: True if LVS tool loading was successful, False otherwise.
        """
        config_result = self.instantiate_tool_from_config("lvs", HammerLVSTool)
        if config_result is None:
            return False
        else:
            (lvs_tool, name) = config_result
            assert isinstance(lvs_tool, HammerLVSTool)
            return self.set_up_lvs_tool(lvs_tool, name, run_dir)

    def set_up_sram_generator_tool(self, sram_generator_tool: HammerSRAMGeneratorTool,
                        name: str, run_dir: str = "") -> bool:
        """
        Set up and store the given SRAM Generator tool instance for use in this
        driver.
        :param lvs_tool: Tool instance.
        :param name: Short name (e.g. "sram_generator") of the tool instance. Typically
                     obtained from the database.
        :param run_dir: Directory to use for the tool run_dir. Defaults to the
                        run_dir passed in the HammerDriver constructor.
        :return: True if setup was successful.
        """
        if self.tech is None:
            self.log.error("Must load technology before loading SRAM Generator tool")
            return False

        if run_dir == "":
            run_dir = os.path.join(self.obj_dir, "sram_generator-rundir")

        sram_generator_tool.name = name
        sram_generator_tool.logger = self.log.context("sram_generator")
        sram_generator_tool.technology = self.tech
        sram_generator_tool.set_database(self.database)
        sram_generator_tool.submit_command = HammerSubmitCommand.get("sram_generator", self.database)
        sram_generator_tool.run_dir = run_dir
        raw_params = self.database.get_setting("vlsi.inputs.sram_parameters",nullvalue=[])
        sram_params = list(map(lambda p: SRAMParameters.from_setting(p), raw_params))
        sram_generator_tool.input_parameters = sram_params
        # TODO: support hierarchical?

        if len(sram_generator_tool.input_parameters) == 0:
            self.log.warning("No SRAM parameters specified, no SRAMs will be generated.")

        self.sram_generator_tool = sram_generator_tool
        self.tool_configs["sram_generator"], self.tool_config_types["sram_generator"] = sram_generator_tool.get_config()
        self.update_tool_configs()
        return True

    def load_sram_generator_tool(self, run_dir: str = "") -> bool:
        """
        Loads an SRAM Generator tool on a given database

        :param run_dir: Directory to use for the tool run_dir. Defaults to the run_dir passed in the HammerDriver
                        constructor.
        :return: True if SRAM Generator tool loading was successful, False otherwise.
        """
        config_result = self.instantiate_tool_from_config("sram_generator", HammerSRAMGeneratorTool)
        if config_result is None:
            return False
        else:
            (sram_generator_tool, name) = config_result
            assert isinstance(sram_generator_tool, HammerSRAMGeneratorTool)
            return self.set_up_sram_generator_tool(sram_generator_tool, name, run_dir)


    def set_up_sim_tool(self, sim_tool: HammerSimTool,
                              name: str, run_dir: str = "") -> bool:
        """
        Set up and store the given simulation tool instance for use in this
        driver.
        :param sim_tool: Tool instance.
        :param name: Short name (e.g. "vcs") of the tool instance. Typically
                     obtained from the database.
        :param run_dir: Directory to use for the tool run_dir. Defaults to the
                        run_dir passed in the HammerDriver constructor.
        :return: True if setup was successful.
        """

        if self.tech is None:
            self.log.error("Must load technology before loading sim tool")
            return False

        if run_dir == "":
            run_dir = os.path.join(self.obj_dir, "sim-rundir")

        sim_tool.name = name
        sim_tool.logger = self.log.context("sim")
        sim_tool.set_database(self.database)
        sim_tool.run_dir = run_dir
        sim_tool.technology = self.tech
        sim_tool.input_files = self.database.get_setting("sim.inputs.input_files")
        sim_tool.top_module = self.database.get_setting("sim.inputs.top_module", nullvalue="")
        sim_tool.hierarchical_mode = HierarchicalMode.from_str(
            self.database.get_setting("vlsi.inputs.hierarchical.mode"))
        # Special case: if non-leaf hierarchical and gate-level, append ilm sim netlists
        if sim_tool.hierarchical_mode.is_nonleaf_hierarchical() and sim_tool.level.is_gatelevel():
            for ilm in sim_tool.get_input_ilms():
                if isinstance(ilm.sim_netlist, str):
                    sim_tool.input_files.append(ilm.sim_netlist)
        sim_tool.input_files = self.database.get_setting("sim.inputs.input_files")
        sim_tool.submit_command = HammerSubmitCommand.get("sim", self.database)
        sim_tool.all_regs = self.database.get_setting("sim.inputs.all_regs")
        sim_tool.seq_cells = self.database.get_setting("sim.inputs.seq_cells")
        sim_tool.sdf_file = self.database.get_setting("sim.inputs.sdf_file")

        missing_inputs = False
        if sim_tool.top_module == "":
            self.log.error("Top module not specified for simulation")
            missing_inputs = True
        if len(sim_tool.input_files) == 0:
            self.log.error("No input files specified for simulation")
            missing_inputs = True
        if missing_inputs:
            return False

        self.sim_tool = sim_tool
        self.tool_configs["simulation"], self.tool_config_types["simulation"] = sim_tool.get_config()
        self.update_tool_configs()
        return True

    def load_sim_tool(self, run_dir: str = "") -> bool:
        """
        Load the simulation tool based on the given database.

        :param run_dir: Directory to use for the tool run_dir. Defaults to the run_dir passed in the HammerDriver
                        constructor.
        :return: True if simulation tool loading was successful, False otherwise.
        """
        config_result = self.instantiate_tool_from_config("sim", HammerSimTool)
        if config_result is None:
            return False
        else:
            (sim_tool, name) = config_result
            assert isinstance(sim_tool, HammerSimTool)
            return self.set_up_sim_tool(sim_tool, name, run_dir)

    def set_up_power_tool(self, power_tool: HammerPowerTool,
                              name: str, run_dir: str = "") -> bool:
        """
        Set up and store the given power tool instance for use in this
        driver.
        :param power_tool: Tool instance.
        :param name: Short name (e.g. "vcs") of the tool instance. Typically
                     obtained from the database.
        :param run_dir: Directory to use for the tool run_dir. Defaults to the
                        run_dir passed in the HammerDriver constructor.
        :return: True if setup was successful.
        """
        if self.tech is None:
            self.log.error("Must load technology before loading power tool")
            return False

        if run_dir == "":
            run_dir = os.path.join(self.obj_dir, "power-rundir")

        power_tool.name = name
        power_tool.logger = self.log.context("power")
        power_tool.technology = self.tech
        power_tool.set_database(self.database)
        power_tool.hierarchical_mode = HierarchicalMode.from_str(
            self.database.get_setting("vlsi.inputs.hierarchical.mode"))
        power_tool.submit_command = HammerSubmitCommand.get("power", self.database)
        power_tool.run_dir = run_dir

        power_tool.flow_database = self.database.get_setting("power.inputs.database", nullvalue="")
        power_tool.spefs = self.database.get_setting("power.inputs.spefs", nullvalue=[])
        power_tool.waveforms = self.database.get_setting("power.inputs.waveforms", nullvalue=[])
        power_tool.saifs = self.database.get_setting("power.inputs.saifs", nullvalue=[])

        power_tool.input_files = self.database.get_setting("power.inputs.input_files", nullvalue=[])
        power_tool.sdc = self.database.get_setting("power.inputs.sdc", nullvalue="")
        power_tool.top_module = self.database.get_setting("power.inputs.top_module", nullvalue="")
        power_tool.tb_name = self.database.get_setting("power.inputs.tb_name", nullvalue="")
        power_tool.tb_dut = self.database.get_setting("power.inputs.tb_dut", nullvalue="")

        missing_inputs = False
        if power_tool.flow_database == "" and len(power_tool.input_files) == 0:
            self.log.error("Input database or design not specified for power analysis")
            missing_inputs = True

        if missing_inputs:
            return False

        self.power_tool = power_tool
        self.tool_configs["power"], self.tool_config_types["power"] = power_tool.get_config()
        self.update_tool_configs()
        return True

    def load_power_tool(self, run_dir: str = "") -> bool:
        """
        Loads a power tool on a given database

        :param run_dir: Directory to use for the tool run_dir. Defaults to the run_dir passed in the HammerDriver
                        constructor.
        :return: True if power tool loading was successful, False otherwise.
        """
        config_result = self.instantiate_tool_from_config("power", HammerPowerTool)
        if config_result is None:
            return False
        else:
            (power_tool, name) = config_result
            assert isinstance(power_tool, HammerPowerTool)
            return self.set_up_power_tool(power_tool, name, run_dir)

    def set_up_formal_tool(self, formal_tool: HammerFormalTool,
                              name: str, run_dir: str = "") -> bool:
        """
        Set up and store the given formal tool instance for use in this
        driver.
        :param formal_tool: Tool instance.
        :param name: Short name (e.g. "conformal") of the tool instance. Typically
                     obtained from the database.
        :param run_dir: Directory to use for the tool run_dir. Defaults to the
                        run_dir passed in the HammerDriver constructor.
        :return: True if setup was successful.
        """

        if self.tech is None:
            self.log.error("Must load technology before loading formal tool")
            return False

        if run_dir == "":
            run_dir = os.path.join(self.obj_dir, "formal-rundir")

        formal_tool.name = name
        formal_tool.logger = self.log.context("formal")
        formal_tool.set_database(self.database)
        formal_tool.run_dir = run_dir
        formal_tool.technology = self.tech
        formal_tool.submit_command = HammerSubmitCommand.get("formal", self.database)

        formal_tool.check = self.database.get_setting("formal.inputs.check")
        formal_tool.input_files = self.database.get_setting("formal.inputs.input_files")
        formal_tool.hierarchical_mode = HierarchicalMode.from_str(
            self.database.get_setting("vlsi.inputs.hierarchical.mode"))
        # Special case: if non-leaf hierarchical, append ilm sim netlists
        if formal_tool.hierarchical_mode.is_nonleaf_hierarchical():
            for ilm in formal_tool.get_input_ilms():
                if isinstance(ilm.sim_netlist, str):
                    formal_tool.input_files.append(ilm.sim_netlist)
        formal_tool.reference_files = self.database.get_setting("formal.inputs.reference_files")
        formal_tool.top_module = self.database.get_setting("formal.inputs.top_module")
        missing_inputs = False
        if formal_tool.check == "":
            self.log.error("No check specified for formal")
            missing_inputs = True
        if len(formal_tool.input_files) == 0:
            self.log.error("No input files specified for formal")
            missing_inputs = True
        if missing_inputs:
            return False

        self.formal_tool = formal_tool

        self.tool_configs["formal"], self.tool_config_types["formal"] = formal_tool.get_config()
        self.update_tool_configs()
        return True

    def load_formal_tool(self, run_dir: str = "") -> bool:
        """
        Loads a formal tool on a given database

        :param run_dir: Directory to use for the tool run_dir. Defaults to the run_dir passed in the HammerDriver
                        constructor.
        :return: True if formal tool loading was successful, False otherwise.
        """
        config_result = self.instantiate_tool_from_config("formal", HammerFormalTool)
        if config_result is None:
            return False
        else:
            (formal_tool, name) = config_result
            assert isinstance(formal_tool, HammerFormalTool)
            return self.set_up_formal_tool(formal_tool, name, run_dir)

    def set_up_timing_tool(self, timing_tool: HammerTimingTool,
                              name: str, run_dir: str = "") -> bool:
        """
        Set up and store the given timing tool instance for use in this
        driver.
        :param timing_tool: Tool instance.
        :param name: Short name (e.g. "tempus") of the tool instance. Typically
                     obtained from the database.
        :param run_dir: Directory to use for the tool run_dir. Defaults to the
                        run_dir passed in the HammerDriver constructor.
        :return: True if setup was successful.
        """

        if self.tech is None:
            self.log.error("Must load technology before loading timing tool")
            return False

        if run_dir == "":
            run_dir = os.path.join(self.obj_dir, "timing-rundir")

        timing_tool.name = name
        timing_tool.logger = self.log.context("timing")
        timing_tool.set_database(self.database)
        timing_tool.run_dir = run_dir
        timing_tool.technology = self.tech
        timing_tool.submit_command = HammerSubmitCommand.get("timing", self.database)

        timing_tool.input_files = self.database.get_setting("timing.inputs.input_files")
        timing_tool.hierarchical_mode = HierarchicalMode.from_str(
            self.database.get_setting("vlsi.inputs.hierarchical.mode"))
        timing_tool.top_module = self.database.get_setting("timing.inputs.top_module")
        missing_inputs = False
        if len(timing_tool.input_files) == 0:
            self.log.error("No input files specified for timing")
            missing_inputs = True
        if missing_inputs:
            return False

        self.timing_tool = timing_tool

        self.tool_configs["timing"], self.tool_config_types["timing"] = timing_tool.get_config()
        self.update_tool_configs()
        return True

    def load_timing_tool(self, run_dir: str = "") -> bool:
        """
        Loads a timing tool on a given database

        :param run_dir: Directory to use for the tool run_dir. Defaults to the run_dir passed in the HammerDriver
                        constructor.
        :return: True if timing tool loading was successful, False otherwise.
        """
        config_result = self.instantiate_tool_from_config("timing", HammerTimingTool)
        if config_result is None:
            return False
        else:
            (timing_tool, name) = config_result
            assert isinstance(timing_tool, HammerTimingTool)
            return self.set_up_timing_tool(timing_tool, name, run_dir)

    def set_up_pcb_tool(self, pcb_tool: HammerPCBDeliverableTool,
                              name: str, run_dir: str = "") -> bool:
        """
        Set up and store the given PCB deliverable tool instance for use in this
        driver.
        :param pcb_tool: Tool instance.
        :param name: Short name (e.g. "pcb") of the tool instance. Typically
                     obtained from the database.
        :param run_dir: Directory to use for the tool run_dir. Defaults to the
                        run_dir passed in the HammerDriver constructor.
        :return: True if setup was successful.
        """
        if self.tech is None:
            self.log.error("Must load technology before loading PCB deliverable tool")
            return False

        if run_dir == "":
            run_dir = os.path.join(self.obj_dir, "pcb-rundir")

        pcb_tool.name = name
        pcb_tool.logger = self.log.context("pcb")
        pcb_tool.technology = self.tech
        pcb_tool.set_database(self.database)
        pcb_tool.top_module = self.database.get_setting("pcb.inputs.top_module", nullvalue="")
        if pcb_tool.top_module == "":
            self.log.error("Top module not specified for PCB")
            return False
        pcb_tool.submit_command = HammerSubmitCommand.get("pcb", self.database)
        pcb_tool.run_dir = run_dir

        self.pcb_tool = pcb_tool
        self.tool_configs["pcb"], self.tool_config_types["pcb"] = pcb_tool.get_config()
        self.update_tool_configs()
        return True

    def load_pcb_tool(self, run_dir: str = "") -> bool:
        """
        Load the PCB deliverable tool based on the given database.

        :param run_dir: Directory to use for the tool run_dir. Defaults to the run_dir passed in the HammerDriver
                        constructor.
        :return: True if successful, false otherwise
        """
        config_result = self.instantiate_tool_from_config("pcb", HammerPCBDeliverableTool)
        if config_result is None:
            return False
        else:
            (pcb_tool, name) = config_result
            assert isinstance(pcb_tool, HammerPCBDeliverableTool)
            return self.set_up_pcb_tool(pcb_tool, name, run_dir)

    def set_post_custom_syn_tool_hooks(self, hooks: List[HammerToolHookAction]) -> None:
        """
        Set the extra list of hooks used for control flow (resume/pause) in run_synthesis.
        They will run after main/hook_actions.

        :param hooks: Hooks to run
        """
        self.post_custom_syn_tool_hooks = list(hooks)

    def set_post_custom_par_tool_hooks(self, hooks: List[HammerToolHookAction]) -> None:
        """
        Set the extra list of hooks used for control flow (resume/pause) in run_par.
        They will run after main/hook_actions.

        :param hooks: Hooks to run
        """
        self.post_custom_par_tool_hooks = list(hooks)

    def set_post_custom_drc_tool_hooks(self, hooks: List[HammerToolHookAction]) -> None:
        """
        Set the extra list of hooks used for control flow (resume/pause) in run_drc.
        They will run after main/hook_actions.

        :param hooks: Hooks to run
        """
        self.post_custom_drc_tool_hooks = list(hooks)

    def set_post_custom_lvs_tool_hooks(self, hooks: List[HammerToolHookAction]) -> None:
        """
        Set the extra list of hooks used for control flow (resume/pause) in run_lvs.
        They will run after main/hook_actions.

        :param hooks: Hooks to run
        """
        self.post_custom_lvs_tool_hooks = list(hooks)

    def set_post_custom_sim_tool_hooks(self, hooks: List[HammerToolHookAction]) -> None:
        """
        Set the extra list of hooks used for control flow (resume/pause) in run_sim.
        They will run after main/hook_actions.

        :param hooks: Hooks to run
        """
        self.post_custom_sim_tool_hooks = list(hooks)

    def set_post_custom_power_tool_hooks(self, hooks: List[HammerToolHookAction]) -> None:
        """
        Set the extra list of hooks used for control flow (resume/pause) in run_power.
        They will run after main/hook_actions.

        :param hooks: Hooks to run
        """
        self.post_custom_power_tool_hooks = list(hooks)

    def set_post_custom_formal_tool_hooks(self, hooks: List[HammerToolHookAction]) -> None:
        """
        Set the extra list of hooks used for control flow (resume/pause) in run_formal.
        They will run after main/hook_actions.

        :param hooks: Hooks to run
        """
        self.post_custom_formal_tool_hooks = list(hooks)

    def set_post_custom_timing_tool_hooks(self, hooks: List[HammerToolHookAction]) -> None:
        """
        Set the extra list of hooks used for control flow (resume/pause) in run_timing.
        They will run after main/hook_actions.

        :param hooks: Hooks to run
        """
        self.post_custom_timing_tool_hooks = list(hooks)

    def run_synthesis(self, hook_actions: Optional[List[HammerToolHookAction]] = None, force_override: bool = False) -> \
            Tuple[bool, dict]:
        """
        Run synthesis based on the given database.
        The output config dict returned does NOT have a copy of the input config settings.

        :param hook_actions: List of hook actions, or leave as None to use the hooks sets in set_synthesis_hooks.
                             Hooks from set_synthesis_hooks, if present, will be appended afterwards.
        :param force_override: Set to true to overwrite instead of append.
        :return: Tuple of (success, output config dict)
        """
        if self.syn_tool is None:
            self.log.error("Must load synthesis tool before calling run_synthesis")
            return False, {}

        # TODO: think about artifact storage?
        self.log.info("Starting synthesis with tool '%s'" % (self.syn_tool.name))
        if hook_actions is None:
            hooks_to_use = self.post_custom_syn_tool_hooks
        else:
            if force_override:
                hooks_to_use = hook_actions
            else:
                hooks_to_use = hook_actions + self.post_custom_syn_tool_hooks
        run_succeeded = self.syn_tool.run(hooks_to_use)
        if not run_succeeded:
            self.log.error("Synthesis tool %s failed! Please check its output." % self.syn_tool.name)
            # Allow the flow to keep running, just in case.
            # TODO: make this an option

        # Record output from the tool into the JSON output.
        # Note: the output config dict is NOT complete
        output_config = {}  # type: Dict[str, Any]
        # TODO(edwardw): automate this
        try:
            output_config = deepdict(self.syn_tool.export_config_outputs())
            if output_config.get("vlsi.builtins.is_complete", True):
                self.log.error(
                    "The synthesis plugin is mis-written; "
                    "it did not mark its output dictionary as output-only "
                    "or did not call super().export_config_outputs(). "
                    "Subsequent commands might not behave correctly.")
                output_config["vlsi.builtins.is_complete"] = False
        except ValueError as e:
            self.log.fatal(e.args[0])
            return False, {}

        return run_succeeded, output_config

    @staticmethod
    def synthesis_output_to_par_input(output_dict: dict) -> Optional[dict]:
        """
        Generate the appropriate inputs for running place-and-route from the
        outputs of synthesis run.
        Does not merge the results with any project dictionaries.
        :param output_dict: Dict containing synthesis.outputs.*
        :return: par.inputs.* settings generated from output_dict,
                 or None if output_dict was invalid
        """
        try:
            output_files = deeplist(output_dict["synthesis.outputs.output_files"])
            result = {
                "par.inputs.input_files": output_files,
                "par.inputs.top_module": output_dict["synthesis.inputs.top_module"],
                "vlsi.builtins.is_complete": False
            }  # type: Dict[str, Any]
            if "synthesis.outputs.sdc" in output_dict:
                result["par.inputs.post_synth_sdc"] = output_dict["synthesis.outputs.sdc"]
            return result
        except KeyError:
            # KeyError means that the given dictionary is missing output keys.
            return None

    @staticmethod
    def synthesis_output_to_sim_input(output_dict: dict) -> Optional[dict]:
        """
        Generate the appropriate inputs for running gate level simulations from the
        outputs of synthesis run.
        Does not merge the results with any project dictionaries.
        :param output_dict: Dict containing synthesis.outputs.*
        :return: sim.inputs.* settings generated from output_dict,
                 or None if output_dict was invalid
        """
        try:
            output_files = deeplist(output_dict["synthesis.outputs.output_files"])
            result = {
                "sim.inputs.input_files": output_files,
                "sim.inputs.input_files_meta": "append",
                "sim.inputs.top_module": output_dict["synthesis.inputs.top_module"],
                "sim.inputs.all_regs": output_dict["synthesis.outputs.all_regs"],
                "sim.inputs.seq_cells": output_dict["synthesis.outputs.seq_cells"],
                "sim.inputs.sdf_file": output_dict["synthesis.outputs.sdf_file"],
                "sim.inputs.level": 'syn',
                "vlsi.builtins.is_complete": False
            }  # type: Dict[str, Any]
            return result
        except KeyError:
            # KeyError means that the given dictionary is missing output keys.
            return None

    @staticmethod
    def synthesis_output_to_power_input(output_dict: dict) -> Optional[dict]:
        """
        Generate the appropriate inputs for running post-synthesis power analysis from the
        outputs of synthesis run.
        Does not merge the results with any project dictionaries.
        :param output_dict: Dict containing synthesis.outputs.*
        :return: power.inputs.* settings generated from output_dict,
                 or None if output_dict was invalid
        """
        try:
            output_files = deeplist(output_dict["synthesis.outputs.output_files"])
            result = {
                "power.inputs.input_files": output_files,
                "power.inputs.input_files_meta": "append",
                "power.inputs.top_module": output_dict["synthesis.inputs.top_module"],
                "power.inputs.level": 'syn',
                "power.inputs.sdc": output_dict["synthesis.outputs.sdc"],
                "vlsi.builtins.is_complete": False
            }  # type: Dict[str, Any]
            return result
        except KeyError:
            # KeyError means that the given dictionary is missing output keys.
            return None

    @staticmethod
    def synthesis_output_to_formal_input(output_dict: dict) -> Optional[dict]:
        """
        Generate the appropriate inputs for running formal tools from the
        outputs of synthesis run.
        Does not merge the results with any project dictionaries.
        :param output_dict: Dict containing synthesis.outputs.*
        :return: formal.inputs.* settings generated from output_dict,
                 or None if output_dict was invalid
        """
        try:
            reference_files = deeplist(output_dict["synthesis.inputs.input_files"])
            input_files = deeplist(output_dict["synthesis.outputs.output_files"])
            result = {
                "formal.inputs.input_files": input_files,
                "formal.inputs.input_files_meta": "append",
                "formal.inputs.reference_files": reference_files,
                "formal.inputs.reference_files_meta": "append",
                "formal.inputs.top_module": output_dict["synthesis.inputs.top_module"],
                "vlsi.builtins.is_complete": False
            }  # type: Dict[str, Any]
            return result
        except KeyError:
            # KeyError means that the given dictionary is missing output keys.
            return None

    @staticmethod
    def synthesis_output_to_timing_input(output_dict: dict) -> Optional[dict]:
        """
        Generate the appropriate inputs for running timing tools from the
        outputs of synthesis run.
        Does not merge the results with any project dictionaries.
        :param output_dict: Dict containing synthesis.outputs.*
        :return: timing.inputs.* settings generated from output_dict,
                 or None if output_dict was invalid
        """
        try:
            input_files = deeplist(output_dict["synthesis.outputs.output_files"])
            result = {
                "timing.inputs.input_files": input_files,
                "timing.inputs.input_files_meta": "append",
                "timing.inputs.top_module": output_dict["synthesis.inputs.top_module"],
                "timing.inputs.post_synth_sdc": output_dict["synthesis.outputs.sdc"],
                "timing.inputs.sdf_file": output_dict["synthesis.outputs.sdf_file"],
                "vlsi.builtins.is_complete": False
            }  # type: Dict[str, Any]
            return result
        except KeyError:
            # KeyError means that the given dictionary is missing output keys.
            return None

    @staticmethod
    def par_output_to_sim_input(output_dict: dict) -> Optional[dict]:
        """
        Generate the appropriate inputs for running gate level simulations from the
        outputs of par run.
        Does not merge the results with any project dictionaries.
        :param output_dict: Dict containing par.outputs.*
        :return: sim.inputs.* settings generated from output_dict,
                 or None if output_dict was invalid
        """
        try:
            sim_input_files = deeplist([output_dict["par.outputs.output_sim_netlist"]])
            result = {
                "sim.inputs.input_files": sim_input_files,
                "sim.inputs.input_files_meta": "append",
                "sim.inputs.top_module": output_dict["par.inputs.top_module"],
                "sim.inputs.all_regs": output_dict["par.outputs.all_regs"],
                "sim.inputs.seq_cells": output_dict["par.outputs.seq_cells"],
                "sim.inputs.sdf_file": output_dict["par.outputs.sdf_file"],
                "sim.inputs.level": 'par',
                "vlsi.builtins.is_complete": False
            }  # type: Dict[str, Any]
            return result
        except KeyError:
            # KeyError means that the given dictionary is missing output keys.
            return None

    @staticmethod
    def par_output_to_power_input(output_dict: dict) -> Optional[dict]:
        """
        Generate the appropriate inputs for running power analysis from the
        outputs of par run.
        Does not merge the results with any project dictionaries.
        :param output_dict: Dict containing par.outputs.*
        :return: power.inputs.* settings generated from output_dict,
                 or None if output_dict was invalid
        """
        try:
            result = {
                "power.inputs.top_module": output_dict["par.inputs.top_module"],
                "power.inputs.netlist": output_dict["par.outputs.output_netlist"],
                "power.inputs.spefs": output_dict["par.outputs.spefs"],
                "vlsi.builtins.is_complete": False
            }  # type: Dict[str, Any]
            return result
        except KeyError:
            # KeyError means that the given dictionary is missing output keys.
            return None

    @staticmethod
    def par_output_to_formal_input(output_dict: dict) -> Optional[dict]:
        """
        Generate the appropriate inputs for running formal tools from the
        outputs of par run.
        Does not merge the results with any project dictionaries.
        :param output_dict: Dict containing par.outputs.*
        :return: formal.inputs.* settings generated from output_dict,
                 or None if output_dict was invalid
        """
        try:
            reference_files = deeplist(output_dict["par.inputs.input_files"])
            input_files = deeplist([output_dict["par.outputs.output_sim_netlist"]])
            result = {
                "formal.inputs.input_files": input_files,
                "formal.inputs.input_files_meta": "append",
                "formal.inputs.reference_files": reference_files,
                "formal.inputs.reference_files_meta": "append",
                "formal.inputs.top_module": output_dict["par.inputs.top_module"],
                "vlsi.builtins.is_complete": False
            }  # type: Dict[str, Any]
            return result
        except KeyError:
            # KeyError means that the given dictionary is missing output keys.
            return None

    @staticmethod
    def par_output_to_timing_input(output_dict: dict) -> Optional[dict]:
        """
        Generate the appropriate inputs for running timing tools from the
        outputs of par run.
        Does not merge the results with any project dictionaries.
        :param output_dict: Dict containing par.outputs.*
        :return: timing.inputs.* settings generated from output_dict,
                 or None if output_dict was invalid
        """
        try:
            input_files = deeplist([output_dict["par.outputs.output_netlist"]])
            result = {
                "timing.inputs.input_files": input_files,
                "timing.inputs.input_files_meta": "append",
                "timing.inputs.top_module": output_dict["par.inputs.top_module"],
                "timing.inputs.spefs": output_dict["par.outputs.spefs"],
                "timing.inputs.sdf_file": output_dict["par.outputs.sdf_file"],
                "timing.inputs.ilms": output_dict["par.outputs.output_ilms"],
                "vlsi.builtins.is_complete": False
            }  # type: Dict[str, Any]
            return result
        except KeyError:
            # KeyError means that the given dictionary is missing output keys.
            return None

    def run_par(self, hook_actions: Optional[List[HammerToolHookAction]] = None, force_override: bool = False) -> Tuple[
        bool, dict]:
        """
        Run place and route based on the given database.
        The output config dict returned does NOT have a copy of the input config settings.
        """
        if self.par_tool is None:
            self.log.error("Must load par tool before calling run_par")
            return False, {}

        # TODO: update API to match run_synthesis and deduplicate logic
        self.log.info("Starting place and route with tool '%s'" % (self.par_tool.name))
        if hook_actions is None:
            hooks_to_use = self.post_custom_par_tool_hooks
        else:
            if force_override:
                hooks_to_use = hook_actions
            else:
                hooks_to_use = hook_actions + self.post_custom_par_tool_hooks
        # TODO: get place and route working
        run_succeeded = self.par_tool.run(hooks_to_use)
        if not run_succeeded:
            self.log.error("Place and route tool %s failed! Please check its output." % self.par_tool.name)
            # Allow the flow to keep running, just in case.
            # TODO: make this an option

        # Record output from the tool into the JSON output.
        # Note: the output config dict is NOT complete
        output_config = {}  # type: Dict[str, Any]
        # TODO(edwardw): automate this
        try:
            output_config = deepdict(self.par_tool.export_config_outputs())
            if output_config.get("vlsi.builtins.is_complete", True):
                self.log.error(
                    "The place-and-route plugin is mis-written; "
                    "it did not mark its output dictionary as output-only "
                    "or did not call super().export_config_outputs(). "
                    "Subsequent commands might not behave correctly.")
                output_config["vlsi.builtins.is_complete"] = False
        except ValueError as e:
            self.log.fatal(e.args[0])
            return False, {}

        return run_succeeded, output_config

    @staticmethod
    def par_output_to_syn_input(output_dict: dict) -> Optional[dict]:
        """
        Generate the appropriate inputs for running the next level of synthesis from the
        outputs of par run in a hierarchical flow.
        Does not merge the results with any project dictionaries.
        :param output_dict: Dict containing par.outputs.*
        :return: vlsi.inputs.* settings generated from output_dict,
                 or None if output_dict was invalid
        """
        try:
            result = {
                "vlsi.inputs.ilms": output_dict["par.outputs.output_ilms"],
                "vlsi.builtins.is_complete": False
            }  # type: Dict[str, Any]
            return result
        except KeyError:
            # KeyError means that the given dictionary is missing output keys.
            return None

    @staticmethod
    def par_output_to_drc_input(output_dict: dict) -> Optional[dict]:
        """
        Generate the appropriate inputs for running DRC from the
        outputs of par run.
        Does not merge the results with any project dictionaries.
        :param output_dict: Dict containing par.outputs.*
        :return: drc.inputs.* settings generated from output_dict,
                 or None if output_dict was invalid
        """
        try:
            result = {
                "drc.inputs.top_module": output_dict["par.inputs.top_module"],
                "drc.inputs.layout_file": output_dict["par.outputs.output_gds"],
                "vlsi.builtins.is_complete": False
            }  # type: Dict[str, Any]
            return result
        except KeyError:
            # KeyError means that the given dictionary is missing output keys.
            return None

    @staticmethod
    def par_output_to_lvs_input(output_dict: dict) -> Optional[dict]:
        """
        Generate the appropriate inputs for running LVS from the
        outputs of par run.
        Does not merge the results with any project dictionaries.
        :param output_dict: Dict containing par.outputs.*
        :return: lvs.inputs.* settings generated from output_dict,
                 or None if output_dict was invalid
        """
        try:
            result = {
                "lvs.inputs.top_module": output_dict["par.inputs.top_module"],
                "lvs.inputs.layout_file": output_dict["par.outputs.output_gds"],
                "lvs.inputs.schematic_files": [output_dict["par.outputs.output_netlist"]],
                "lvs.inputs.ilms": output_dict["par.outputs.output_ilms"],
                "lvs.inputs.hcells_list": output_dict["par.outputs.hcells_list"],
                "vlsi.builtins.is_complete": False
            }  # type: Dict[str, Any]
            return result
        except KeyError:
            # KeyError means that the given dictionary is missing output keys.
            return None

    def run_drc(self, hook_actions: Optional[List[HammerToolHookAction]] = None, force_override: bool = False) -> Tuple[
        bool, dict]:
        """
        Run DRC on a given database.

        :param hook_actions: List of hook actions, or leave as None to use the hooks sets in set_drc_hooks.
                             Hooks from set_drc_hooks, if present, will be appended afterwards.
        :param force_override: Set to true to overwrite instead of append.
        :return: Tuple of (success, output config dict)
        """
        if self.drc_tool is None:
            self.log.error("Must load DRC tool before calling run_drc")
            return False, {}

        self.log.info("Starting DRC check with tool '%s'" % (self.drc_tool.name))

        if hook_actions is None:
            hooks_to_use = self.post_custom_drc_tool_hooks
        elif force_override:
            hooks_to_use = hook_actions
        else:
            hooks_to_use = hook_actions + self.post_custom_drc_tool_hooks

        run_succeeded = self.drc_tool.run(hooks_to_use)
        if not run_succeeded:
            self.log.error("DRC tool %s failed! Please check its output." % self.drc_tool.name)
            # Allow the flow to keep running, just in case

        # Record output from the drc_tool into the JSON output
        output_config = {}  # type: Dict[str, Any]
        try:
            output_config.update(self.drc_tool.export_config_outputs())
        except ValueError as e:
            self.log.fatal(e.args[0])
            return False, {}

        return run_succeeded, output_config

    def run_lvs(self, hook_actions: Optional[List[HammerToolHookAction]] = None, force_override: bool = False) -> Tuple[
        bool, dict]:
        """
        Run LVS on a given database.

        :param hook_actions: List of hook actions, or leave as None to use the hooks sets in set_lvs_hooks.
                             Hooks from set_lvs_hooks, if present, will be appended afterwards.
        :param force_override: Set to true to overwrite instead of append.
        :return: Tuple of (success, output config dict)
        """
        if self.lvs_tool is None:
            self.log.error("Must load LVS tool before calling run_lvs")
            return False, {}

        self.log.info("Starting LVS check with tool '%s'" % (self.lvs_tool.name))

        if hook_actions is None:
            hooks_to_use = self.post_custom_lvs_tool_hooks
        elif force_override:
            hooks_to_use = hook_actions
        else:
            hooks_to_use = hook_actions + self.post_custom_lvs_tool_hooks

        run_succeeded = self.lvs_tool.run(hooks_to_use)
        if not run_succeeded:
            self.log.error("LVS tool %s failed! Please check its output." % self.lvs_tool.name)
            # Allow the flow to keep running, just in case

        # Record output from the lvs_tool into the JSON output
        output_config = {}  # type: Dict[str, Any]
        try:
            output_config.update(self.lvs_tool.export_config_outputs())
        except ValueError as e:
            self.log.fatal(e.args[0])
            return False, {}

        return run_succeeded, output_config

    def run_sram_generator(self, hook_actions: Optional[List[HammerToolHookAction]] = None, force_override: bool = False) -> Tuple[
        bool, dict]:
        """
        Run SRAM Generator on a given database.

        :param hook_actions: List of hook actions, or leave as None to use the hooks sets in set_sram_generator_hooks.
                             Hooks from set_sram_generator_hooks, if present, will be appended afterwards.
        :param force_override: Set to true to overwrite instead of append.
        :return: Tuple of (success, output config dict)
        """
        if self.sram_generator_tool is None:
            self.log.error("Must load SRAM Generator tool before calling run_sram_generator")
            return False, {}

        self.log.info("Starting SRAM Generator with tool '%s'" % (self.sram_generator_tool.name))

        if hook_actions is None:
            hooks_to_use = self.post_custom_sram_generator_tool_hooks
        elif force_override:
            hooks_to_use = hook_actions
        else:
            hooks_to_use = hook_actions + self.post_custom_sram_generator_tool_hooks

        run_succeeded = self.sram_generator_tool.run(hooks_to_use)
        if not run_succeeded:
            self.log.error("SRAM Generator tool %s failed! Please check its output." % self.sram_generator_tool.name)
            # Allow the flow to keep running, just in case

        # Record output from the sram_generator_tool into the JSON output
        output_config = {}  # type: Dict[str, Any]
        try:
            output_config.update(self.sram_generator_tool.export_config_outputs())
        except ValueError as e:
            self.log.fatal(e.args[0])
            return False, {}

        return run_succeeded, output_config

    @staticmethod
    def sim_output_to_power_input(output_dict: dict) -> Optional[dict]:
        """
        Generate the inputs for dynamic power analysis from the
        outputs of simulations.
        Does not merge the results with any project dictionaries.
        :param output_dict: Dict containing sim.outputs.*
        :return: sim.inputs.* settings generated from output_dict,
                 or None if output_dict was invalid
        """
        try:
            result = {
                "power.inputs.waveforms": output_dict["sim.outputs.waveforms"],
                "power.inputs.waveforms_meta": "append",
                "power.inputs.saifs": output_dict["sim.outputs.saifs"],
                "power.inputs.saifs_meta": "append",
                "power.inputs.top_module": output_dict["sim.outputs.output_top_module"],
                "power.inputs.tb_name": output_dict["sim.outputs.output_tb_name"],
                "power.inputs.tb_dut": output_dict["sim.outputs.output_tb_dut"],
                "power.inputs.level": output_dict["sim.outputs.output_level"],
                "vlsi.builtins.is_complete": False
            }  # type: Dict[str, Any]
            return result
        except KeyError:
            # KeyError means that the given dictionary is missing output keys.
            return None

    def run_sim(self, hook_actions: Optional[List[HammerToolHookAction]] = None, force_override: bool = False) -> \
            Tuple[bool, dict]:
        """
        Run simulation based on the given database.
        The output config dict returned does NOT have a copy of the input config settings.

        :param hook_actions: List of hook actions, or leave as None to use the hooks sets in set_simulation_hooks.
                             Hooks from set_simulation_hooks, if present, will be appended afterwards.
        :param force_override: Set to true to overwrite instead of append.
        :return: Tuple of (success, output config dict)
        """
        if self.sim_tool is None:
            self.log.error("Must load simulation tool before calling run_sim")
            return False, {}

        # TODO: think about artifact storage?
        self.log.info("Starting simulation with tool '%s'" % (self.sim_tool.name))
        if hook_actions is None:
            hooks_to_use = self.post_custom_sim_tool_hooks
        else:
            if force_override:
                hooks_to_use = hook_actions
            else:
                hooks_to_use = hook_actions + self.post_custom_sim_tool_hooks

        run_succeeded = self.sim_tool.run(hooks_to_use)
        if not run_succeeded:
            self.log.error("Simulation tool %s failed! Please check its output." % self.sim_tool.name)
            # Allow the flow to keep running, just in case.
            # TODO: make this an option

        # Record output from the tool into the JSON output.
        # Note: the output config dict is NOT complete
        output_config = {}  # type: Dict[str, Any]
        try:
            output_config = deepdict(self.sim_tool.export_config_outputs())
            if output_config.get("vlsi.builtins.is_complete", True):
                self.log.error(
                    "The simulation plugin is mis-written; "
                    "it did not mark its output dictionary as output-only "
                    "or did not call super().export_config_outputs(). "
                    "Subsequent commands might not behave correctly.")
                output_config["vlsi.builtins.is_complete"] = False
        except ValueError as e:
            self.log.fatal(e.args[0])
            return False, {}

        return run_succeeded, output_config

    def run_power(self, hook_actions: Optional[List[HammerToolHookAction]] = None, force_override: bool = False) -> \
            Tuple[bool, dict]:
        """
        Run power analysis based on the given database.
        The output config dict returned does NOT have a copy of the input config settings.

        :param hook_actions: List of hook actions, or leave as None to use the hooks sets in set_power_hooks.
                             Hooks from set_power_hooks, if present, will be appended afterwards.
        :param force_override: Set to true to overwrite instead of append.
        :return: Tuple of (success, output config dict)
        """
        if self.power_tool is None:
            self.log.error("Must load power tool before calling run_power")
            return False, {}

        # TODO: think about artifact storage?
        self.log.info("Starting power with tool '%s'" % (self.power_tool.name))
        if hook_actions is None:
            hooks_to_use = self.post_custom_power_tool_hooks
        else:
            if force_override:
                hooks_to_use = hook_actions
            else:
                hooks_to_use = hook_actions + self.post_custom_power_tool_hooks

        run_succeeded = self.power_tool.run(hooks_to_use)
        if not run_succeeded:
            self.log.error("Power tool %s failed! Please check its output." % self.power_tool.name)
            # Allow the flow to keep running, just in case.
            # TODO: make this an option

        # Record output from the power tool into the JSON output.
        output_config = {}  # type: Dict[str, Any]
        try:
            output_config.update(self.power_tool.export_config_outputs())
        except ValueError as e:
            self.log.fatal(e.args[0])
            return False, {}

        return run_succeeded, output_config

    def run_formal(self, hook_actions: Optional[List[HammerToolHookAction]] = None, force_override: bool = False) -> \
            Tuple[bool, dict]:
        """
        Run formal verification based on the given database.
        The output config dict returned does NOT have a copy of the input config settings.

        :param hook_actions: List of hook actions, or leave as None to use the hooks sets in set_formal_hooks.
                             Hooks from set_formal_hooks, if present, will be appended afterwards.
        :param force_override: Set to true to overwrite instead of append.
        :return: Tuple of (success, output config dict)
        """
        if self.formal_tool is None:
            self.log.error("Must load formal tool before calling run_formal")
            return False, {}

        # TODO: think about artifact storage?
        self.log.info("Starting formal with tool '%s'" % (self.formal_tool.name))
        if hook_actions is None:
            hooks_to_use = self.post_custom_formal_tool_hooks
        else:
            if force_override:
                hooks_to_use = hook_actions
            else:
                hooks_to_use = hook_actions + self.post_custom_formal_tool_hooks

        run_succeeded = self.formal_tool.run(hooks_to_use)
        if not run_succeeded:
            self.log.error("Formal tool %s failed! Please check its output." % self.formal_tool.name)
            # Allow the flow to keep running, just in case.
            # TODO: make this an option

        # Record output from the formal tool into the JSON output.
        output_config = {}  # type: Dict[str, Any]
        try:
            output_config.update(self.formal_tool.export_config_outputs())
        except ValueError as e:
            self.log.fatal(e.args[0])
            return False, {}

        return run_succeeded, output_config

    def run_timing(self, hook_actions: Optional[List[HammerToolHookAction]] = None, force_override: bool = False) -> \
            Tuple[bool, dict]:
        """
        Run timing analysis based on the given database.
        The output config dict returned does NOT have a copy of the input config settings.

        :param hook_actions: List of hook actions, or leave as None to use the hooks sets in set_timing_hooks.
                             Hooks from set_timing_hooks, if present, will be appended afterwards.
        :param force_override: Set to true to overwrite instead of append.
        :return: Tuple of (success, output config dict)
        """
        if self.timing_tool is None:
            self.log.error("Must load timing tool before calling run_timing")
            return False, {}

        # TODO: think about artifact storage?
        self.log.info("Starting timing with tool '%s'" % (self.timing_tool.name))
        if hook_actions is None:
            hooks_to_use = self.post_custom_timing_tool_hooks
        else:
            if force_override:
                hooks_to_use = hook_actions
            else:
                hooks_to_use = hook_actions + self.post_custom_timing_tool_hooks

        run_succeeded = self.timing_tool.run(hooks_to_use)
        if not run_succeeded:
            self.log.error("Timing tool %s failed! Please check its output." % self.timing_tool.name)
            # Allow the flow to keep running, just in case.
            # TODO: make this an option

        # Record output from the timing tool into the JSON output.
        output_config = {}  # type: Dict[str, Any]
        try:
            output_config.update(self.timing_tool.export_config_outputs())
        except ValueError as e:
            self.log.fatal(e.args[0])
            return False, {}

        return run_succeeded, output_config

    def run_pcb(self, hook_actions: Optional[List[HammerToolHookAction]] = None, force_override: bool = False) -> Tuple[
        bool, dict]:
        """
        Run the PCB deliverable generation tool

        :param hook_actions: List of hook actions, or leave as None to use the hooks sets in set_pcb_hooks.
                             Hooks from set_pcb_hooks, if present, will be appended afterwards.
        :param force_override: Set to true to overwrite instead of append.
        :return: Tuple of (success, output config dict)
        """
        if self.pcb_tool is None:
            self.log.error("Must load PCB deliverable tool before calling run_pcb")
            return False, {}

        self.log.info("Starting PCB deliverable generation with tool '%s'" % (self.pcb_tool.name))

        if hook_actions is None:
            hooks_to_use = self.post_custom_pcb_tool_hooks
        elif force_override:
            hooks_to_use = hook_actions
        else:
            hooks_to_use = hook_actions + self.post_custom_pcb_tool_hooks

        run_succeeded = self.pcb_tool.run(hooks_to_use)
        if not run_succeeded:
            self.log.error("PCB deliverable tool %s failed! Please check its output." % self.pcb_tool.name)
            # Allow the flow to keep running, just in case

        # Record output from the pcb_tool into the JSON output
        output_config = {}  # type: Dict[str, Any]
        try:
            output_config.update(self.pcb_tool.export_config_outputs())
        except ValueError as e:
            self.log.fatal(e.args[0])
            return False, {}

        return run_succeeded, output_config

    def get_hierarchical_dependency_graph(self) -> Dict[str, Tuple[List[str], List[str]]]:
        """
        Return the dependency graph for this flow if it is hierarchical, or an empty dict if it is flat.
        The flow is the set of setps configured by the current input Hammer IR.

        :return: The dependency graph.
        """
        return self._hierarchical_helper()[1]

    def get_hierarchical_settings(self) -> List[Tuple[str, dict]]:
        """
        Read settings from the database, determine leaf/hierarchical modules, an order of execution, and return an
        ordered list (from leaf to top) of modules and associated config snippets needed to run syn+par for that module
        hierarchically.

        :return: List of tuples of (module name, config snippet)
        """
        return self._hierarchical_helper()[0]

    def _hierarchical_helper(self) -> Tuple[List[Tuple[str, dict]], Dict[str, Tuple[List[str], List[str]]]]:
        """
        Read settings from the database, determine leaf/hierarchical modules, an order of execution, and return an
        ordered list (from leaf to top) of modules and associated config snippets needed to run syn+par for that module
        hierarchically and the dependency graph. Do not call this method directly- use get_hierarchical_settings or
        get_hierarchial_dependency_graph instead.

        :return: Tuple of (List of tuples of (module name, config snippet), the dependency graph)
        """
        hier_source_key = "vlsi.inputs.hierarchical.config_source"
        hier_source = str(self.database.get_setting(hier_source_key))
        hier_modules = {}  # type: Dict[str, List[str]]
        hier_placement_constraints = {}  # type: Dict[str, List[PlacementConstraint]]
        hier_constraints = {}  # type: Dict[str, List[Dict]]

        # This is retrieving the list of hard macro sizes to be used when creating PlacementConstraint tuples later
        list_of_hard_macros = self.database.get_setting("vlsi.technology.extra_macro_sizes")  # type: List[Dict]
        hard_macros = list(map(MacroSize.from_setting, list_of_hard_macros))

        if hier_source == "none":
            pass
        elif hier_source == "manual":
            list_of_hier_modules = self.database.get_setting(
                "vlsi.inputs.hierarchical.manual_modules")  # type: List[Dict]
            assert isinstance(list_of_hier_modules, list)
            if len(list_of_hier_modules) == 0:
                raise ValueError("No hierarchical modules defined manually in manual hierarchical mode")
            hier_modules = reduce(add_dicts, list_of_hier_modules)

            list_of_placement_constraints = self.database.get_setting(
                "vlsi.inputs.hierarchical.manual_placement_constraints")  # type: List[Dict]
            assert isinstance(list_of_placement_constraints, list)
            combined_raw_placement_dict = reduce(add_dicts, list_of_placement_constraints, {})  # type: Dict[str, List[Dict[str, Any]]]

            # This helper function filters only the dict containing the toplevel placement constraint, if any, from the provided list of dicts.
            # If the list does not contain a toplevel constraint, it returns None.
            def get_toplevel(d: List[Dict[str, Any]]) -> Optional[Dict[str, Any]]:
                results = list(filter(lambda x: x["type"] == "toplevel", d))
                if len(results) == 0:
                    return None
                else:
                    return results[-1]

            # Use the above helper method to filter down the combined raw placement dict into a dict:
            # - keys are hierarchical module name
            # - values are dicts containing toplevel constraints or None
            toplevels_opt = {k: get_toplevel(v) for k, v in combined_raw_placement_dict.items()}  # type: Dict[str, Optional[Dict[str, Any]]]
            # This filters out all of the Nones to get only hierarchical modules with toplevel placement constraints
            toplevels = {k: v for k, v in toplevels_opt.items() if v is not None}  # type: Dict[str, Dict[str, Any]]
            # This converts each dict entry into a MacroSize tuple, which should now represent all hierarchical modules
            hier_macros = [MacroSize(library="", name=x[0], width=x[1]["width"], height=x[1]["height"]) for x in toplevels.items()]
            masters = hard_macros + hier_macros

            hier_placement_constraints = {key: list(map(partial(PlacementConstraint.from_masters_and_dict, masters), lst))
                                          for key, lst in combined_raw_placement_dict.items()}
            # Iterate over project configs to find which ones contain hierarchical constraints
            # For each file that does append its path to the special key in the extracted
            # hierarchical constraints section.
            """
            hier_constraint_source = ""  # type: str
            for project_conf in self.project_configs:
                if "vlsi.inputs.hierarchical.constraints" in project_conf:
                    hier_constraint_source = project_conf[hammer_config._CONFIG_PATH_KEY]
                    pc = project_conf["vlsi.inputs.hierarchical.constraints"]  # type: List[Dict]
                    # Add CONFIG_PATH_KEY to actual project configs for each project config's hierarchical constraint
                    # keys then update project configs at the end
                    for md in pc:
                        md  # one entry dict with a list
                        for m in md.keys():
                            if hammer_config._CONFIG_PATH_KEY in md[m][-1]:
                                pass
                            else:
                                md[m].append({hammer_config._CONFIG_PATH_KEY: hier_constraint_source})
            """
            self.update_project_configs(self.project_configs)
            list_of_hier_constraints = self.database.get_setting(
                    "vlsi.inputs.hierarchical.constraints") # type: List[Dict]
            hier_constraints = reduce(add_dicts, list_of_hier_constraints, {})
        elif hier_source == "from_placement":
            raise NotImplementedError("Generation from placement not implemented yet")
        else:
            raise ValueError("Invalid value for " + hier_source_key)

        assert isinstance(hier_modules, dict)
        if not hier_modules:
            return ([], {})

        leaf_modules = set()  # type: Set[str]
        intermediate_modules = set()  # type: Set[str]
        top_module = str(self.database.get_setting("vlsi.inputs.hierarchical.top_module"))
        if top_module == "" or top_module is None:
            raise ValueError("Cannot have a hierarchical flow if the top module is not set")

        # Node + outgoing edges (nodes that depend on us) + incoming edges (nodes we depend on)
        dependency_graph = {}  # type: Dict[str, Tuple[List[str], List[str]]]

        # If there is a hierarchy, find the leaf and intermediate modules.
        def visit_module(mod: str) -> None:
            if mod not in hier_modules:
                if mod == top_module:
                    raise ValueError("Cannot have a hierarchical flow with top as leaf")
                leaf_modules.add(mod)
                return
            elif len(hier_modules[mod]) == 0:
                if mod == top_module:
                    raise ValueError("Cannot have a hierarchical flow with top as leaf")
                leaf_modules.add(mod)
                return
            else:
                if mod != top_module:
                    intermediate_modules.add(mod)
                for m in hier_modules[mod]:
                    # m depends on us
                    dependency_graph.setdefault(m, ([], []))[0].append(mod)
                    # We depend on m
                    dependency_graph.setdefault(mod, ([], []))[1].append(m)
                    visit_module(m)
        visit_module(top_module)

        # Create an order for the modules to be run in.
        order = topological_sort(dependency_graph, list(leaf_modules))

        output = []  # type: List[Tuple[str, dict]]

        for module in order:
            mode = HierarchicalMode.Hierarchical
            if module == top_module:
                mode = HierarchicalMode.Top
            elif module in leaf_modules:
                mode = HierarchicalMode.Leaf
            elif module in intermediate_modules:
                mode = HierarchicalMode.Hierarchical
            else:
                assert "Should not get here"

            constraint_dict = {
                "vlsi.inputs.hierarchical.mode": str(mode),
                "synthesis.inputs.top_module": module,
                "vlsi.inputs.placement_constraints": list(
                    map(PlacementConstraint.to_dict, hier_placement_constraints.get(module, [])))
            }
            constraint_dict = reduce(add_dicts, hier_constraints.get(module, []), constraint_dict)
            output.append((module, constraint_dict))

        return (output, dependency_graph)

    @property
    def dump_history(self) -> bool:
        return self._dump_history
    
    @dump_history.setter
    def dump_history(self, new_dump_history: bool) -> None:
        self._dump_history = new_dump_history



# Yosys synthesis plugin for Hammer (as part of OpenROAD-flow installation).
# mostly adoped from OpenLANE scripts found here: https://github.com/The-OpenROAD-Project/OpenLane/tree/master/scripts
#
# See LICENSE for licence details.

from hammer.vlsi.vendor import OpenROADTool, OpenROADSynthesisTool

from hammer.vlsi import HammerTool, HammerToolStep, HammerToolHookAction, HierarchicalMode, TCLTool
from hammer.vlsi.constraints import MMMCCorner, MMMCCornerType
from hammer.utils import VerilogUtils, optional_map
from hammer.vlsi import HammerSynthesisTool
from hammer.logging import HammerVLSILogging
from hammer.vlsi import MMMCCornerType
import hammer.tech as hammer_tech

from typing import Dict, List, Any, Optional, Callable
import importlib.resources
from pathlib import Path

import hammer.tech.specialcells as specialcells
from hammer.tech.specialcells import CellType, SpecialCell

import os
import json
from collections import Counter

class YosysSynth(HammerSynthesisTool, OpenROADTool, TCLTool):

    #=========================================================================
    # overrides from parent classes
    #=========================================================================
    @property
    def steps(self) -> List[HammerToolStep]:
        return self.make_steps_from_methods([
            self.init_environment,
            self.syn_generic,
            self.syn_map,
            self.add_tieoffs,
            self.write_regs,
            self.generate_reports,
            self.write_outputs
        ])

    def do_post_steps(self) -> bool:
        assert super().do_post_steps()
        return self.run_yosys()

    @property
    def env_vars(self) -> Dict[str, str]:
        new_dict = dict(super().env_vars)
        new_dict["YOSYS_BIN"] = self.get_setting("synthesis.yosys.yosys_bin")
        return new_dict

    @property
    def post_synth_sdc(self) -> Optional[str]:
        # No post-synth SDC input for synthesis...
        return None

    @property
    def all_regs_path(self) -> str:
        return os.path.join(self.run_dir, "find_regs_paths.json")

    @property
    def all_cells_path(self) -> str:
        return os.path.join(self.run_dir, "find_regs_cells.json")

    def fill_outputs(self) -> bool:
        # TODO: actually generate the following for simulation
        # Check that the regs paths were written properly if the write_regs step was run
        self.output_seq_cells = self.all_cells_path
        self.output_all_regs = self.all_regs_path
        if self.ran_write_regs:
            if not os.path.isfile(self.all_cells_path):
                raise ValueError("Output find_regs_cells.json %s not found" % (self.all_cells_path))

            if not os.path.isfile(self.all_regs_path):
                raise ValueError("Output find_regs_paths.json %s not found" % (self.all_regs_path))

            # if not self.process_reg_paths(self.all_regs_path):
            #     self.logger.error("Failed to process all register paths")
        else:
            self.logger.info("Did not run write_regs")

        # Check that the synthesis outputs exist if the synthesis run was successful
        mapped_v = self.mapped_v_path
        self.output_files = [mapped_v]
        self.output_sdc = self.mapped_sdc_path
        # self.sdf_file = self.output_sdf_path
        if self.ran_write_outputs:
            if not os.path.isfile(mapped_v):
                raise ValueError("Output mapped verilog %s not found" % (mapped_v)) # better error?

            if not os.path.isfile(self.mapped_sdc_path):
                raise ValueError("Output SDC %s not found" % (self.mapped_sdc_path)) # better error?

        else:
            self.logger.info("Did not run write_outputs")

        return True

    def export_config_outputs(self) -> Dict[str, Any]:
        outputs = dict(super().export_config_outputs())
        outputs["synthesis.outputs.sdc"] = self.output_sdc
        outputs["synthesis.outputs.seq_cells"] = self.output_seq_cells
        outputs["synthesis.outputs.all_regs"] = self.output_all_regs
        outputs["synthesis.outputs.sdf_file"] = ""
        return outputs

    def tool_config_prefix(self) -> str:
        return "synthesis.yosys"

    #=========================================================================
    # useful subroutines
    #=========================================================================
    @property
    def mapped_v_path(self) -> str:
        return os.path.join(self.run_dir, "{}.mapped.v".format(self.top_module))

    @property
    def mapped_sdc_path(self) -> str:
        return os.path.join(self.run_dir, "{}.mapped.sdc".format(self.top_module))

    @property
    def mapped_blif_path(self) -> str:
        return os.path.join(self.run_dir, "{}.mapped.blif".format(self.top_module))

    def get_timing_libs(self, corner: Optional[MMMCCorner] = None) -> str:
        """
        Helper function to get the list of ASCII timing .lib files in space separated format.

        :param corner: Optional corner to consider. If supplied, this will use filter_for_mmmc to select libraries that
        match a given corner (voltage/temperature).
        :return: List of lib files separated by spaces
        """
        pre_filters = optional_map(corner, lambda c: [self.filter_for_mmmc(voltage=c.voltage,
                                                                           temp=c.temp)])  # type: Optional[List[Callable[[hammer_tech.Library],bool]]]

        lib_args = self.technology.read_libs([hammer_tech.filters.timing_lib_with_ecsm_filter],
                                             hammer_tech.HammerTechnologyUtils.to_plain_item,
                                             extra_pre_filters=pre_filters)

        # lib files are often in a zipped format, so unzip them if they are
        lib_args_unzipped = self.technology.extract_gz_files(lib_args)

        return " ".join(lib_args_unzipped)

    def run_yosys(self) -> bool:
        """Close out the synthesis script and run Yosys."""
        # Quit Yosys.
        self.append("exit")

        # Create synthesis script.
        syn_tcl_filename = os.path.join(self.run_dir, "syn.tcl")

        with open(syn_tcl_filename, "w") as f:
            f.write("\n".join(self.output))

        # Build args.
        args = [
            self.get_setting("synthesis.yosys.yosys_bin"),
            "-c", syn_tcl_filename,
        ]

        if bool(self.get_setting("synthesis.yosys.generate_only")):
            self.logger.info("Generate-only mode: command-line is " + " ".join(args))
        else:
            # Temporarily disable colours/tag to make run output more readable.
            # TODO: think of a more elegant way to do this?
            HammerVLSILogging.enable_colour = False
            HammerVLSILogging.enable_tag = False
            self.run_executable(args, cwd=self.run_dir) # TODO: check for errors and deal with them
            HammerVLSILogging.enable_colour = True
            HammerVLSILogging.enable_tag = True

        return True

    def write_sdc_file(self) -> bool:
        with open(self.mapped_sdc_path,'w') as f:
            # custom sdc constraints
            if self.driver_cell is not None:
                f.write(f"set_driving_cell {self.driver_cell}\n" )
                # TODO: generate this hard-coded 5
                f.write("set_load 5\n")
        return True

    #========================================================================
    # synthesis main steps
    #========================================================================
    def init_environment(self) -> bool:

        # set variables to match global variables in OpenLANE
        time_unit = self.get_time_unit().value_prefix + self.get_time_unit().unit
        clock_port = self.get_clock_ports()[0]
        self.clock_port_name = clock_port.name
        time_unit = "ps" # yosys requires time units in ps
        self.clock_period = int(clock_port.period.value_in_units(time_unit))
        self.clock_uncertainty = int(clock_port.period.value_in_units(time_unit))
        self.clock_transition = 0.15 # SYNTH_CLOCK_TRANSITION

        self.synth_cap_load = 33.5 # SYNTH_CAP_LOAD
        self.max_fanout = 5 # default SYNTH_MAX_FANOUT = 5


        self.driver_cell = None
        driver_cells = self.technology.get_special_cell_by_type(CellType.Driver)
        if len(driver_cells) == 0:
            self.logger.warning("Driver cells are unspecified and will not be added during synthesis.")
        elif driver_cells[0].input_ports is None or driver_cells[0].output_ports is None:
            self.logger.warning("Driver cell input and output ports are unspecified and will not be added during synthesis.")
        else:
            self.driver_cell = driver_cells[0].name[0]
            self.driver_ports_in = driver_cells[0].input_ports[0]
            self.driver_ports_out = driver_cells[0].output_ports[0]

        # Yosys commands only take a single lib file
        #   so use typical corner liberty file
        corners = self.get_mmmc_corners()  # type: List[MMMCCorner]
        try:
            corner_tt = next((corner for corner in corners if corner.type == MMMCCornerType.Extra), None)
        except:
            raise ValueError("An extra corner is required for Yosys.")

        self.liberty_files_tt = self.get_timing_libs(corner_tt)

        self.append("yosys -import")

        # replaces undef (x) constants with defined (0/1) constants.
        self.append("setundef -zero")

        # We are switching working directories and Yosys still needs to find paths.
        abspath_input_files = list(map(lambda name: os.path.join(os.getcwd(), name), self.input_files))  # type: List[str]

        # Add any verilog_synth wrappers (which are needed in some technologies e.g. for SRAMs) which need to be
        # synthesized.
        abspath_input_files += self.technology.read_libs([
            hammer_tech.filters.verilog_synth_filter
        ], hammer_tech.HammerTechnologyUtils.to_plain_item)

        for verilog_file in abspath_input_files:
            self.append(f"read_verilog -sv {verilog_file}")

        liberty_files = self.technology.read_libs([hammer_tech.filters.timing_lib_with_ecsm_filter], hammer_tech.HammerTechnologyUtils.to_plain_item)
        for lib_file in liberty_files:
            self.append(f"read_liberty -lib {lib_file}")

        return True

    def syn_generic(self) -> bool:
        # TODO: is there a better way to do this? like self.get_setting()
        if self._database.has_setting("synthesis.yosys.latch_map_file"):
            latch_map = f"techmap -map {self.get_setting('synthesis.yosys.latch_map_file')}"
        else:  # TODO: make the else case better
            latch_map = ""

        self.block_append(f"""
        # TODO: verify this command, it was in yosys manual but not in OpenLANE script
        yosys proc
        hierarchy -check -top {self.top_module}

        synth -top {self.top_module}

        # Optimize the design
        opt -purge

        # Technology mapping of latches
        {latch_map}

        # Technology mapping of flip-flops
        """)
        for liberty_file in self.liberty_files_tt.split():
            self.verbose_append(f"dfflibmap -liberty {liberty_file}")
        self.verbose_append("opt")

        self.write_sdc_file()
        return True

    def syn_map(self) -> bool:

        self.block_append(f"""
        # Technology mapping for cells
        # ABC supports multiple liberty files, but the hook from Yosys to ABC doesn't
        # TODO: this is a bad way of getting one liberty file, need a way to merge all std cell lib files
        abc -D {self.clock_period} \\
            -constr "{self.mapped_sdc_path}" \\
            -liberty "{self.liberty_files_tt.split()[0]}" \\
            -showtmp

        # Replace undef values with defined constants
        # TODO: do we need this??
        setundef -zero

        # Split multi-bit nets into single-bit nets.
        # Splitting nets resolves unwanted compound assign statements in netlist (assign [..] = [..])
        splitnets

        # Remove unused cells and wires
        opt_clean -purge
        """)
        return True

    def add_tieoffs(self) -> bool:
        tie_hi_cells = self.technology.get_special_cell_by_type(CellType.TieHiCell)
        tie_lo_cells = self.technology.get_special_cell_by_type(CellType.TieLoCell)
        tie_hilo_cells = self.technology.get_special_cell_by_type(CellType.TieHiLoCell)

        if len(tie_hi_cells) != 1 or len (tie_lo_cells) != 1 or tie_hi_cells[0].output_ports is None or tie_lo_cells[0].output_ports is None:
            self.logger.warning("Hi and Lo tiecells and their input ports are unspecified or improperly specified and will not be added during synthesis.")
        else:
            tie_hi_cell = tie_hi_cells[0].name[0]
            tie_hi_port = tie_hi_cells[0].output_ports[0]
            tie_lo_cell = tie_lo_cells[0].name[0]
            tie_lo_port = tie_lo_cells[0].output_ports[0]

            self.block_append(f"""
            # Technology mapping of constant hi- and/or lo-drivers
            hilomap -singleton \\
                    -hicell {{*}}{tie_hi_cell} {tie_hi_port} \\
                    -locell {{*}}{tie_lo_cell} {tie_lo_port}
            """)
        if self.driver_cell is not None:
            self.block_append(f"""
            # Insert driver cells for pass through wires
            insbuf -buf {{*}}{self.driver_cell} {self.driver_ports_in} {self.driver_ports_out}
            """)
        return True

    def write_regs(self) -> bool:
        # TODO: generate find_regs_cells.json / find_regs_paths.json here
        self.ran_write_regs = False
        return True

    def generate_reports(self) -> bool:
        # TODO: generate all reports (will probably need to parse the log file)
        self.block_append(f"""
        tee -o {self.run_dir}/{self.top_module}.synth_check.rpt check

        tee -o {self.run_dir}/{self.top_module}.synth_stat.txt stat -top {self.top_module} -liberty {self.liberty_files_tt.split()[0]}
        """)
        return True

    def write_outputs(self) -> bool:
        hier_mapped_v_path=os.path.join(self.run_dir, f"{self.top_module}.mapped.hier.v")
        self.block_append(f"""
        write_verilog -noattr -noexpr -nohex -nodec -defparam "{self.mapped_v_path}"

        # flatten

        # # OpenROAD will throw an error if the verilog from Yosys is not flattened
        # # UPDATE ON THIS: nvm, it somehow works now...
        # write_verilog -noattr -noexpr -nohex -nodec -defparam "{self.mapped_v_path}"

        # BLIF file seems to be easier to parse than mapped verilog for find_regs functions so leave for now
        # write_blif -top {self.top_module} "{self.mapped_blif_path}"
        """)
        self.ran_write_outputs = True
        return True

tool = YosysSynth



#  hammer-vlsi plugin for Cadence Innovus.
#
#  See LICENSE for licence details.

import shutil
from typing import List, Dict, Optional, Callable, Tuple, Any, cast

import os
import errno

from hammer.utils import get_or_else, optional_map
from hammer.vlsi import HammerTool, HammerPlaceAndRouteTool, HammerToolStep, HammerToolHookAction, \
    PlacementConstraintType, HierarchicalMode, ILMStruct, ObstructionType, Margins, Supply, PlacementConstraint, MMMCCornerType
from hammer.vlsi.units import CapacitanceValue
from hammer.logging import HammerVLSILogging
import hammer.tech as hammer_tech
from hammer.tech import RoutingDirection
from hammer.tech.specialcells import CellType
from decimal import Decimal
from hammer.common.cadence import CadenceTool

# Notes: camelCase commands are the old syntax (deprecated)
# snake_case commands are the new/common UI syntax.
# This plugin should only use snake_case commands.

class Innovus(HammerPlaceAndRouteTool, CadenceTool):

    def export_config_outputs(self) -> Dict[str, Any]:
        outputs = dict(super().export_config_outputs())
        # TODO(edwardw): find a "safer" way of passing around these settings keys.
        outputs["par.outputs.seq_cells"] = self.output_seq_cells
        outputs["par.outputs.all_regs"] = self.output_all_regs
        outputs["par.outputs.sdf_file"] = self.output_sdf_path
        outputs["par.outputs.spefs"] = self.output_spef_paths
        return outputs

    def fill_outputs(self) -> bool:
        if self.ran_write_ilm:
            # Check that the ILMs got written.

            ilm_data_dir = "{ilm_dir_name}/mmmc/ilm_data/{top}".format(ilm_dir_name=self.ilm_dir_name,
                                                                       top=self.top_module)
            postRoute_v_gz = os.path.join(ilm_data_dir, "{top}_postRoute.v.gz".format(top=self.top_module))

            if not os.path.isfile(postRoute_v_gz):
                raise ValueError("ILM output postRoute.v.gz %s not found" % (postRoute_v_gz))

            # Copy postRoute.v.gz to postRoute.ilm.v.gz since that's what Genus seems to expect.
            postRoute_ilm_v_gz = os.path.join(ilm_data_dir, "{top}_postRoute.ilm.v.gz".format(top=self.top_module))
            shutil.copyfile(postRoute_v_gz, postRoute_ilm_v_gz)

            # Write output_ilms.
            self.output_ilms = [
                ILMStruct(dir=self.ilm_dir_name, data_dir=ilm_data_dir, module=self.top_module,
                          lef=os.path.join(self.run_dir, "{top}ILM.lef".format(top=self.top_module)),
                          gds=self.output_gds_filename,
                          netlist=self.output_netlist_filename,
                          sim_netlist=self.output_sim_netlist_filename)
            ]
        else:
            self.output_ilms = []

        # Check that the regs paths were written properly if the write_regs step was run
        self.output_seq_cells = self.all_cells_path
        self.output_all_regs = self.all_regs_path
        if self.ran_write_regs:
            if not os.path.isfile(self.all_cells_path):
                raise ValueError("Output find_regs_cells.json %s not found" % (self.all_cells_path))

            if not os.path.isfile(self.all_regs_path):
                raise ValueError("Output find_regs_paths.json %s not found" % (self.all_regs_path))

            if not self.process_reg_paths(self.all_regs_path):
                self.logger.error("Failed to process all register paths")
        else:
            self.logger.info("Did not run write_regs")

        # Check that the par outputs exist if the par run was successful
        self.output_gds = self.output_gds_filename
        self.output_netlist = self.output_netlist_filename
        self.output_sim_netlist = self.output_sim_netlist_filename
        self.hcells_list = []
        self.sdf_file = self.output_sdf_path
        self.spef_files = self.output_spef_paths

        if self.ran_write_design:
            if not os.path.isfile(self.output_gds_filename):
                raise ValueError("Output GDS %s not found" % (self.output_gds_filename))

            if not os.path.isfile(self.output_netlist_filename):
                raise ValueError("Output netlist %s not found" % (self.output_netlist_filename))

            if not os.path.isfile(self.output_sim_netlist_filename):
                raise ValueError("Output sim netlist %s not found" % (self.output_sim_netlist_filename))

            if not os.path.isfile(self.output_sdf_path):
                raise ValueError("Output SDF %s not found" % (self.output_sdf_path))

            for spef_path in self.output_spef_paths:
                if not os.path.isfile(spef_path):
                    raise ValueError("Output SPEF %s not found" % (spef_path))
        else:
            self.logger.info("Did not run write_design")

        return True

    @property
    def output_gds_filename(self) -> str:
        return os.path.join(self.run_dir, "{top}.gds".format(top=self.top_module))

    @property
    def output_netlist_filename(self) -> str:
        return os.path.join(self.run_dir, "{top}.lvs.v".format(top=self.top_module))

    @property
    def output_sim_netlist_filename(self) -> str:
        return os.path.join(self.run_dir, "{top}.sim.v".format(top=self.top_module))

    @property
    def all_regs_path(self) -> str:
        return os.path.join(self.run_dir, "find_regs_paths.json")

    @property
    def all_cells_path(self) -> str:
        return os.path.join(self.run_dir, "find_regs_cells.json")

    @property
    def output_sdf_path(self) -> str:
        return os.path.join(self.run_dir, "{top}.par.sdf".format(top=self.top_module))

    @property
    def output_spef_paths(self) -> List[str]:
        corners = self.get_mmmc_corners()
        if corners:
            # Order matters in tool consuming spefs (ensured here by get_mmmc_corners())!
            return list(map(lambda c: os.path.join(self.run_dir, "{top}.{corner}.par.spef".format(top=self.top_module, corner=c.name)), corners))
        else:
            return [os.path.join(self.run_dir, "{top}.par.spef".format(top=self.top_module))]

    @property
    def env_vars(self) -> Dict[str, str]:
        v = dict(super().env_vars)
        v["INNOVUS_BIN"] = self.get_setting("par.innovus.innovus_bin")
        return v

    @property
    def _step_transitions(self) -> List[Tuple[str, str]]:
        """
        Private helper property to keep track of which steps we ran so that we
        can create symlinks.
        This is a list of (pre, post) steps
        """
        return self.attr_getter("__step_transitions", [])

    @_step_transitions.setter
    def _step_transitions(self, value: List[Tuple[str, str]]) -> None:
        self.attr_setter("__step_transitions", value)

    def do_pre_steps(self, first_step: HammerToolStep) -> bool:
        assert super().do_pre_steps(first_step)
        # Restore from the last checkpoint if we're not starting over.
        if first_step != self.first_step:
            self.verbose_append("read_db pre_{step}".format(step=first_step.name))
        return True

    def do_between_steps(self, prev: HammerToolStep, next: HammerToolStep) -> bool:
        assert super().do_between_steps(prev, next)
        # Write a checkpoint to disk.
        self.verbose_append("write_db pre_{step}".format(step=next.name))
        # Symlink the database to latest for open_chip script later.
        self.verbose_append("ln -sfn pre_{step} latest".format(step=next.name))
        self._step_transitions = self._step_transitions + [(prev.name, next.name)]
        return True

    def do_post_steps(self) -> bool:
        assert super().do_post_steps()
        # Create symlinks for post_<step> to pre_<step+1> to improve usability.
        try:
            for prev, next in self._step_transitions:
                os.symlink(
                    os.path.join(self.run_dir, "pre_{next}".format(next=next)), # src
                    os.path.join(self.run_dir, "post_{prev}".format(prev=prev)) # dst
                )
        except OSError as e:
            if e.errno != errno.EEXIST:
                self.logger.warning("Failed to create post_* symlinks: " + str(e))

        # Create db post_<last step>
        # TODO: this doesn't work if you're only running the very last step
        if len(self._step_transitions) > 0:
            last = "post_{step}".format(step=self._step_transitions[-1][1])
            self.verbose_append("write_db {last}".format(last=last))
            # Symlink the database to latest for open_chip script later.
            self.verbose_append("ln -sfn {last} latest".format(last=last))

        return self.run_innovus()

    def get_tool_hooks(self) -> List[HammerToolHookAction]:
        return [self.make_persistent_hook(innovus_global_settings)]

    @property
    def steps(self) -> List[HammerToolStep]:
        steps = [
            self.init_design,
            self.floorplan_design,
            self.place_bumps,
            self.place_tap_cells,
            self.power_straps,
            self.place_pins,
            self.place_opt_design,
            self.clock_tree,
            self.add_fillers,
            self.route_design,
            self.opt_design
        ]
        write_design_step = [
            self.write_regs,
            self.write_design
        ]  # type: List[Callable[[], bool]]
        if self.hierarchical_mode == HierarchicalMode.Flat:
            # Nothing to do
            pass
        elif self.hierarchical_mode == HierarchicalMode.Leaf:
            # All modules in hierarchical must write an ILM
            write_design_step += [self.write_ilm]
        elif self.hierarchical_mode == HierarchicalMode.Hierarchical:
            # All modules in hierarchical must write an ILM
            write_design_step += [self.write_ilm]
        elif self.hierarchical_mode == HierarchicalMode.Top:
            # No need to write ILM at the top.
            # Top needs assemble_design instead.
            steps += [self.assemble_design]
            pass
        else:
            raise NotImplementedError("HierarchicalMode not implemented: " + str(self.hierarchical_mode))
        return self.make_steps_from_methods(steps + write_design_step)

    def tool_config_prefix(self) -> str:
        return "par.innovus"

    def init_design(self) -> bool:
        """Initialize the design."""
        verbose_append = self.verbose_append

        # Perform common path pessimism removal in setup and hold mode
        verbose_append("set_db timing_analysis_cppr both")
        # Use OCV mode for timing analysis by default
        verbose_append("set_db timing_analysis_type ocv")
        # Match SDC time units to timing libraries
        verbose_append("set_library_unit -time 1{}".format(self.get_time_unit().value_prefix + self.get_time_unit().unit))

        # Read LEF layouts.
        lef_files = self.technology.read_libs([
            hammer_tech.filters.lef_filter
        ], hammer_tech.HammerTechnologyUtils.to_plain_item)
        if self.hierarchical_mode.is_nonleaf_hierarchical():
            ilm_lefs = list(map(lambda ilm: ilm.lef, self.get_input_ilms()))
            lef_files.extend(ilm_lefs)
        verbose_append("read_physical -lef {{ {files} }}".format(
            files=" ".join(lef_files)
        ))

        # Read timing libraries.
        mmmc_path = os.path.join(self.run_dir, "mmmc.tcl")
        self.write_contents_to_path(self.generate_mmmc_script(), mmmc_path)
        verbose_append("read_mmmc {mmmc_path}".format(mmmc_path=mmmc_path))

        # Read netlist.
        # Innovus only supports structural Verilog for the netlist; the Verilog can be optionally compressed.
        if not self.check_input_files([".v", ".v.gz"]):
            return False

        # We are switching working directories and we still need to find paths.
        abspath_input_files = list(map(lambda name: os.path.join(os.getcwd(), name), self.input_files))
        verbose_append("read_netlist {{ {files} }} -top {top}".format(
            files=" ".join(abspath_input_files),
            top=self.top_module
        ))

        if self.hierarchical_mode.is_nonleaf_hierarchical():
            # Read ILMs.
            for ilm in self.get_input_ilms():
                # Assumes that the ILM was created by Innovus (or at least the file/folder structure).
                verbose_append("read_ilm -cell {module} -directory {dir}".format(dir=ilm.dir, module=ilm.module))

        # Emit init_power_nets and init_ground_nets in case CPF/UPF is not used
        # commit_power_intent does not override power nets defined in "init_power_nets"
        spec_mode = self.get_setting("vlsi.inputs.power_spec_mode")  # type: str
        if spec_mode == "empty":
            power_supplies = self.get_independent_power_nets()  # type: List[Supply]
            power_nets = " ".join(map(lambda s: s.name, power_supplies))
            ground_supplies = self.get_independent_ground_nets()  # type: List[Supply]
            ground_nets = " ".join(map(lambda s: s.name, ground_supplies))
            verbose_append("set_db init_power_nets {{{n}}}".format(n=power_nets))
            verbose_append("set_db init_ground_nets {{{n}}}".format(n=ground_nets))

        # Run init_design to validate data and start the Cadence place-and-route workflow.
        verbose_append("init_design")

        # Setup power settings from cpf/upf
        for l in self.generate_power_spec_commands():
            verbose_append(l)

        # Set design effort.
        verbose_append("set_db design_flow_effort {}".format(self.get_setting("par.innovus.design_flow_effort")))

        # Set "don't use" cells.
        for l in self.generate_dont_use_commands():
            self.append(l)

        return True

    def floorplan_design(self) -> bool:
        floorplan_tcl = os.path.join(self.run_dir, "floorplan.tcl")
        self.write_contents_to_path("\n".join(self.create_floorplan_tcl()), floorplan_tcl)
        self.verbose_append("source -echo -verbose {}".format(floorplan_tcl))
        return True

    def place_bumps(self) -> bool:
        bumps = self.get_bumps()
        if bumps is not None:
            bump_array_width = Decimal(str((bumps.x - 1) * bumps.pitch_x))
            bump_array_height = Decimal(str((bumps.y - 1) * bumps.pitch_y))
            fp_consts = self.get_placement_constraints()
            fp_width = Decimal(0)
            fp_height = Decimal(0)
            for const in fp_consts:
                if const.type == PlacementConstraintType.TopLevel:
                    fp_width = const.width
                    fp_height = const.height
            if fp_width == 0 or fp_height == 0:
                raise ValueError("Floorplan does not specify a TopLevel constraint or it has no dimensions")
            # Center bump array in the middle of floorplan
            bump_offset_x = (Decimal(str(fp_width)) - bump_array_width) / 2 + bumps.global_x_offset
            bump_offset_y = (Decimal(str(fp_height)) - bump_array_height) / 2+ bumps.global_y_offset
            power_ground_nets = list(map(lambda x: x.name, self.get_independent_power_nets() + self.get_independent_ground_nets()))
            # TODO: Fix this once the stackup supports vias ucb-bar/hammer#354
            block_layer = self.get_setting("vlsi.technology.bump_block_cut_layer")  # type: str
            for bump in bumps.assignments:
                self.append("create_bump -cell {cell} -location_type cell_center -name_format \"Bump_{c}.{r}\" -orient r0 -location \"{x} {y}\"".format(
                    cell = bump.custom_cell if bump.custom_cell is not None else bumps.cell,
                    c = bump.x,
                    r = bump.y,
                    x = bump_offset_x + Decimal(str(bump.x - 1)) * Decimal(str(bumps.pitch_x)),
                    y = bump_offset_y + Decimal(str(bump.y - 1)) * Decimal(str(bumps.pitch_y))))
                if not bump.no_connect:
                    if bump.name in power_ground_nets:
                        self.append("select_bumps -bumps \"Bump_{x}.{y}\"".format(x=bump.x, y=bump.y))
                        self.append("assign_pg_bumps -selected -nets {n}".format(n=bump.name))
                        self.append("deselect_bumps")
                    else:
                        self.append("assign_signal_to_bump -bumps \"Bump_{x}.{y}\" -net {n}".format(x=bump.x, y=bump.y, n=bump.name))
                self.append("create_route_blockage {layer_options} \"{llx} {lly} {urx} {ury}\"".format(
                    layer_options="-layers {{{l}}} -rects".format(l=block_layer) if(self.version() >= self.version_number("181")) else "-cut_layers {{{l}}} -area".format(l=block_layer),
                    llx = "[get_db bump:Bump_{x}.{y} .bbox.ll.x]".format(x=bump.x, y=bump.y),
                    lly = "[get_db bump:Bump_{x}.{y} .bbox.ll.y]".format(x=bump.x, y=bump.y),
                    urx = "[get_db bump:Bump_{x}.{y} .bbox.ur.x]".format(x=bump.x, y=bump.y),
                    ury = "[get_db bump:Bump_{x}.{y} .bbox.ur.y]".format(x=bump.x, y=bump.y)))
        return True

    def place_tap_cells(self) -> bool:
        tap_cells = self.technology.get_special_cell_by_type(CellType.TapCell)

        if len(tap_cells) == 0:
            self.logger.warning("Tap cells are improperly defined in the tech plugin and will not be added. This step should be overridden with a user hook or tapcell special cell should be added to the tech.json.")
            return True

        tap_cell = tap_cells[0].name[0]

        try:
            interval = self.get_setting("vlsi.technology.tap_cell_interval")
            offset = self.get_setting("vlsi.technology.tap_cell_offset")
            self.append("set_db add_well_taps_cell {TAP_CELL}".format(TAP_CELL=tap_cell))
            self.append("add_well_taps -cell_interval {INTERVAL} -in_row_offset {OFFSET}".format(INTERVAL=interval, OFFSET=offset))
        except KeyError:
            pass
        return True

    def place_pins(self) -> bool:
        fp_consts = self.get_placement_constraints()
        topconst = None  # type: Optional[PlacementConstraint]
        for const in fp_consts:
            if const.type == PlacementConstraintType.TopLevel:
                topconst = const
        if topconst is None:
            self.logger.fatal("Cannot find top-level constraints to place pins")
            return False

        const = cast(PlacementConstraint, topconst)
        assert isinstance(const.margins, Margins), "Margins must be defined for the top level"
        fp_llx = const.margins.left
        fp_lly = const.margins.bottom
        fp_urx = const.width - const.margins.right
        fp_ury = const.height - const.margins.top

        pin_assignments = self.get_pin_assignments()
        self.verbose_append("set_db assign_pins_edit_in_batch true")

        promoted_pins = []  # type: List[str]
        for pin in pin_assignments:
            if pin.preplaced:
                # First set promoted pins
                self.verbose_append("set_promoted_macro_pin -pins {{ {p} }}".format(p=pin.pins))
                promoted_pins.extend(pin.pins.split())
            else:
                # TODO: Do we need pin blockages for our layers?
                # Seems like we will only need special pin blockages if the vias are larger than the straps

                cadence_side = None  # type: Optional[str]
                if pin.side is not None:
                    if pin.side == "internal":
                        cadence_side = "inside"
                    else:
                        cadence_side = pin.side
                side_arg = get_or_else(optional_map(cadence_side, lambda s: "-side " + s), "")

                start_arg = ""
                end_arg = ""
                assign_arg = ""
                pattern_arg = ""

                if pin.location is None:
                    start_arg = "-{start} {{ {sx} {sy} }}".format(
                        start="start" if pin.side == "bottom" or pin.side == "right" else "end",
                        sx=fp_urx if pin.side != "left" else fp_llx,
                        sy=fp_ury if pin.side != "bottom" else fp_lly)

                    end_arg = "-{end} {{ {ex} {ey} }}".format(
                        end="end" if pin.side == "bottom" or pin.side == "right" else "start",
                        ex=fp_llx if pin.side != "right" else fp_urx,
                        ey=fp_lly if pin.side != "top" else fp_ury
                    )
                    if pin.layers and len(pin.layers) > 1:
                        pattern_arg = "-pattern fill_optimised"
                    else:
                        pattern_arg = "-spread_type range"
                else:
                    assign_arg = "-assign {{ {x} {y} }}".format(x=pin.location[0], y=pin.location[1])

                layers_arg = ""
                if pin.layers is not None and len(pin.layers) > 0:
                    layers_arg = "-layer {{ {} }}".format(" ".join(pin.layers))

                width_arg = get_or_else(optional_map(pin.width, lambda f: "-pin_width {f}".format(f=f)), "")
                depth_arg = get_or_else(optional_map(pin.depth, lambda f: "-pin_depth {f}".format(f=f)), "")

                cmd = [
                    "edit_pin",
                    "-fixed_pin",
                    "-pin", pin.pins,
                    "-hinst", self.top_module,
                    pattern_arg,
                    layers_arg,
                    side_arg,
                    start_arg,
                    end_arg,
                    assign_arg,
                    width_arg,
                    depth_arg
                ]

                self.verbose_append(" ".join(cmd))

        # In case the * wildcard is used after preplaced pins, this will place promoted pins correctly.
        # Innovus errors instead of warns if the name matching does not work (e.g. bad wildcards).
        for ppin in promoted_pins:
            self.verbose_append("assign_io_pins -move_fixed_pin -pins [get_db [get_db pins -if {{.name == {p} }}] .net.name]".format(p=ppin))

        self.verbose_append("set_db assign_pins_edit_in_batch false")
        return True

    def power_straps(self) -> bool:
        """Place the power straps for the design."""
        power_straps_tcl = os.path.join(self.run_dir, "power_straps.tcl")
        self.write_contents_to_path("\n".join(self.create_power_straps_tcl()), power_straps_tcl)
        self.verbose_append("source -echo -verbose {}".format(power_straps_tcl))
        return True

    def place_opt_design(self) -> bool:
        """Place the design and do pre-routing optimization."""
        self.verbose_append("place_opt_design")
        return True

    def clock_tree(self) -> bool:
        """Setup and route a clock tree for clock nets."""
        if self.hierarchical_mode.is_nonleaf_hierarchical():
            self.verbose_append('''
            flatten_ilm
            update_constraint_mode -name my_constraint_mode -ilm_sdc_files {sdc}
            '''.format(sdc=self.post_synth_sdc), clean=True)
        if len(self.get_clock_ports()) > 0:
            # Ignore clock tree when there are no clocks
            self.verbose_append("create_clock_tree_spec")
            if bool(self.get_setting("par.innovus.use_cco")):
                # -hold is a secret flag for ccopt_design (undocumented anywhere)
                self.verbose_append("ccopt_design -hold -report_dir hammer_cts_debug -report_prefix hammer_cts")
            else:
                self.verbose_append("clock_design")
        if self.hierarchical_mode.is_nonleaf_hierarchical():
            self.verbose_append("unflatten_ilm")
        return True

    def add_fillers(self) -> bool:
        """add decap and filler cells"""
        decaps = self.technology.get_special_cell_by_type(CellType.Decap)
        stdfillers = self.technology.get_special_cell_by_type(CellType.StdFiller)

        if len(decaps) == 0:
            self.logger.info("The technology plugin 'special cells: decap' field does not exist. It should specify a list of decap cells. Filling with stdfiller instead.")
        else:
            decap_cells = decaps[0].name
            decap_caps = []  # type: List[float]
            if decaps[0].size is not None:
                decap_caps = list(map(lambda x: CapacitanceValue(x).value_in_units("fF"), decaps[0].size))
            if len(decap_cells) != len(decap_caps):
                self.logger.error("Each decap cell in the name list must have a corresponding decapacitance value in the size list.")
            decap_consts = list(filter(lambda x: x.target=="capacitance", self.get_decap_constraints()))
            if len(decap_consts) > 0:
                if decap_caps is None:
                    self.logger.warning("No decap capacitances specified but decap constraints with target: 'capacitance' exist. Add decap capacitances to the tech plugin!")
                else:
                    for (cell, cap) in zip(decap_cells, decap_caps):
                        self.append("add_decap_cell_candidates {CELL} {CAP}".format(CELL=cell, CAP=cap))
                    for const in decap_consts:
                        assert isinstance(const.capacitance, CapacitanceValue)
                        area_str = ""
                        if all(c is not None for c in (const.x, const.y, const.width, const.height)):
                            assert isinstance(const.x, Decimal)
                            assert isinstance(const.y, Decimal)
                            assert isinstance(const.width, Decimal)
                            assert isinstance(const.height, Decimal)
                            area_str = " ".join(("-area", str(const.x), str(const.y), str(const.x+const.width), str(const.y+const.height)))
                        self.verbose_append("add_decaps -effort high -total_cap {CAP} {AREA}".format(
                            CAP=const.capacitance.value_in_units("fF"), AREA=area_str))

        if len(stdfillers) == 0:
            self.logger.warning(
                "The technology plugin 'special cells: stdfiller' field does not exist. It should specify a list of (non IO) filler cells. No filler will be added. You can override this with an add_fillers hook if you do not specify filler cells in the technology plugin.")
        else:
            # Decap cells as fillers
            if len(decaps) > 0:
                fill_cells = list(map(lambda c: str(c), decaps[0].name))
                self.append("set_db add_fillers_cells \"{FILLER}\"".format(FILLER=" ".join(fill_cells)))
                # Targeted decap constraints
                decap_consts = list(filter(lambda x: x.target=="density", self.get_decap_constraints()))
                for const in decap_consts:
                    area_str = ""
                    if all(c is not None for c in (const.x, const.y, const.width, const.height)):
                        assert isinstance(const.x, Decimal)
                        assert isinstance(const.y, Decimal)
                        assert isinstance(const.width, Decimal)
                        assert isinstance(const.height, Decimal)
                        area_str = " ".join(("-area", str(const.x), str(const.y), str(const.x+const.width), str(const.y+const.height)))
                    self.verbose_append("add_fillers -density {DENSITY} {AREA}".format(
                        DENSITY=str(const.density), AREA=area_str))
                # Or, fill everywhere if no decap constraints given
                if len(self.get_decap_constraints()) == 0:
                    self.verbose_append("add_fillers")

            # Then the rest is stdfillers
            fill_cells = list(map(lambda c: str(c), stdfillers[0].name))
            self.append("set_db add_fillers_cells \"{FILLER}\"".format(FILLER=" ".join(fill_cells)))
            self.verbose_append("add_fillers")
        return True


    def route_design(self) -> bool:
        """Route the design."""
        if self.hierarchical_mode.is_nonleaf_hierarchical():
            self.verbose_append("flatten_ilm")

        # Allow express design effort to complete running.
        # By default, route_design will abort in express mode with
        # "WARNING (NRIG-142) Express flow by default will not run routing".
        self.verbose_append("set_db design_express_route true")

        self.verbose_append("route_design")
        return True

    def opt_design(self) -> bool:
        """
        Post-route optimization and fix setup & hold time violations.
        -expanded_views creates timing reports for each MMMC view.
        """
        self.verbose_append("opt_design -post_route -setup -hold -expanded_views")
        if self.hierarchical_mode.is_nonleaf_hierarchical():
            self.verbose_append("unflatten_ilm")
        return True

    def assemble_design(self) -> bool:
        # TODO: implement the assemble_design step.
        return True

    def write_netlist(self) -> bool:
        # Don't use virtual connects (using colon, e.g. VSS:) because they mess up LVS
        self.verbose_append("set_db write_stream_virtual_connection false")

        # Connect power nets that are tied together
        for pwr_gnd_net in (self.get_all_power_nets() + self.get_all_ground_nets()):
            if pwr_gnd_net.tie is not None:
                self.verbose_append("connect_global_net {tie} -type net -net_base_name {net}".format(tie=pwr_gnd_net.tie, net=pwr_gnd_net.name))

        # Output the Verilog netlist for the design and include physical cells (-phys) like decaps and fill
        self.verbose_append("write_netlist {netlist} -top_module_first -top_module {top} -exclude_leaf_cells -phys -flat -exclude_insts_of_cells {{ {pcells} }} ".format(
            netlist=self.output_netlist_filename,
            top=self.top_module,
            pcells=" ".join(self.get_physical_only_cells())
        ))

        self.verbose_append("write_netlist {netlist} -top_module_first -top_module {top} -exclude_leaf_cells -exclude_insts_of_cells {{ {pcells} }} ".format(
            netlist=self.output_sim_netlist_filename,
            top=self.top_module,
            pcells=" ".join(self.get_physical_only_cells())
        ))

        return True

    def write_gds(self) -> bool:
        map_file = get_or_else(
            optional_map(self.get_gds_map_file(), lambda f: "-map_file {}".format(f)),
            ""
        )

        gds_files = self.technology.read_libs([
            hammer_tech.filters.gds_filter
        ], hammer_tech.HammerTechnologyUtils.to_plain_item)
        if self.hierarchical_mode.is_nonleaf_hierarchical():
            ilm_gds = list(map(lambda ilm: ilm.gds, self.get_input_ilms()))
            gds_files.extend(ilm_gds)

        # If we are not merging, then we want to use -output_macros.
        # output_macros means that Innovus should take any macros it has and
        # just output the cells into the GDS. These cells will not have physical
        # information inside them and will need to be merged with some other
        # step later. We do not care about uniquifying them because Innovus will
        # output a single cell for each instance (essentially already unique).

        # On the other hand, if we tell Innovus to do the merge then it is going
        # to get a GDS with potentially multiple child cells and we then tell it
        # to uniquify these child cells in case of name collisons. Without that
        # we could have one child that applies to all cells of that name which
        # is often not what you want.
        # For example, if macro ADC1 has a subcell Comparator which is different
        # from ADC2's subcell Comparator, we don't want ADC1's Comparator to
        # replace ADC2's Comparator.
        # Note that cells not present in the GDSes to be merged will be emitted
        # as-is in the output (like with -output_macros).
        merge_options = "-output_macros" if not self.get_setting("par.inputs.gds_merge") else "-uniquify_cell_names -merge {{ {} }}".format(
            " ".join(gds_files)
        )

        # If the user has specified the par.inputs.gds_precision_mode as
        # "auto", we make write_stream modify the GDS precision according to
        # the value of par.inputs.gds_precision. If this setting is not
        # specified then we fall back to the default behavior, which is to use
        # the units specified in the LEF. See documentation for write_stream
        # for what this switch does and what valid values are.
        unit = ""
        if (self.get_setting("par.inputs.gds_precision_mode") == "manual"):
            gds_precision = self.get_setting("par.inputs.gds_precision") or ""
            valid_values = [100, 200, 1000, 2000, 10000, 20000]
            if gds_precision in valid_values:
                unit = "-unit %s" % gds_precision
            else:
                self.logger.error(
                    "par.inputs.gds_precision value of \"%s\" must be one of %s" %(
                        gds_precision, ', '.join([str(x) for x in valid_values])));
                return False
        # "auto", i.e. not "manual", means not specifying anything extra.

        self.verbose_append(
            "write_stream -mode ALL {map_file} {merge_options} {unit} {gds}".format(
            map_file=map_file,
            merge_options=merge_options,
            gds=self.output_gds_filename,
            unit=unit
        ))
        return True

    def write_sdf(self) -> bool:
        if self.hierarchical_mode.is_nonleaf_hierarchical():
            self.verbose_append("flatten_ilm")

        # Output the Standard Delay Format File for use in timing annotated gate level simulations
        self.verbose_append("write_sdf {run_dir}/{top}.par.sdf".format(run_dir=self.run_dir, top=self.top_module))

        return True

    def write_spefs(self) -> bool:
        # Output a SPEF file that contains the parasitic extraction results
        self.verbose_append("set_db extract_rc_coupled true")
        self.verbose_append("extract_rc")
        corners = self.get_mmmc_corners()
        if corners:
            for corner in corners:
                # Setting up views for all defined corner types: setup, hold, extra
                if corner.type is MMMCCornerType.Setup:
                    corner_type_name = "setup"
                elif corner.type is MMMCCornerType.Hold:
                    corner_type_name = "hold"
                elif corner.type is MMMCCornerType.Extra:
                    corner_type_name = "extra"
                else:
                    raise ValueError("Unsupported MMMCCornerType")

                self.verbose_append("write_parasitics -spef_file {run_dir}/{top}.{cname}.par.spef -rc_corner {cname}.{ctype}_rc".format(run_dir=self.run_dir, top=self.top_module, cname=corner.name, ctype=corner_type_name))

        else:
            self.verbose_append("write_parasitics -spef_file {run_dir}/{top}.par.spef".format(run_dir=self.run_dir, top=self.top_module))

        return True



    @property
    def output_innovus_lib_name(self) -> str:
        return "{top}_FINAL".format(top=self.top_module)

    @property
    def generated_scripts_dir(self) -> str:
        return os.path.join(self.run_dir, "generated-scripts")

    @property
    def open_chip_script(self) -> str:
        return os.path.join(self.generated_scripts_dir, "open_chip")

    @property
    def open_chip_tcl(self) -> str:
        return self.open_chip_script + ".tcl"

    def write_regs(self) -> bool:
        """write regs info to be read in for simulation register forcing"""
        if self.hierarchical_mode.is_nonleaf_hierarchical():
            self.append('flatten_ilm')
            self.append(self.child_modules_tcl())
        self.append(self.write_regs_tcl())
        if self.hierarchical_mode.is_nonleaf_hierarchical():
            self.append('unflatten_ilm')
        self.ran_write_regs = True
        return True

    def write_design(self) -> bool:
        # Save the Innovus design.
        self.verbose_append("write_db {lib_name} -def -verilog".format(
            lib_name=self.output_innovus_lib_name
        ))

        # Write netlist
        self.write_netlist()

        # GDS streamout.
        self.write_gds()

        # Write SDF
        self.write_sdf()

        # Write SPEF
        self.write_spefs()

        # Make sure that generated-scripts exists.
        os.makedirs(self.generated_scripts_dir, exist_ok=True)

        self.ran_write_design=True

        return True

    @property
    def ran_write_regs(self) -> bool:
        """The write_regs step sets this to True if it was run."""
        return self.attr_getter("_ran_write_regs", False)

    @ran_write_regs.setter
    def ran_write_regs(self, val: bool) -> None:
        self.attr_setter("_ran_write_regs", val)

    @property
    def ran_write_design(self) -> bool:
        """The write_design step sets this to True if it was run."""
        return self.attr_getter("_ran_write_design", False)

    @ran_write_design.setter
    def ran_write_design(self, val: bool) -> None:
        self.attr_setter("_ran_write_design", val)

    @property
    def ran_write_ilm(self) -> bool:
        """The write_ilm stage sets this to True if it was run."""
        return self.attr_getter("_ran_write_ilm", False)

    @ran_write_ilm.setter
    def ran_write_ilm(self, val: bool) -> None:
        self.attr_setter("_ran_write_ilm", val)

    @property
    def ilm_dir_name(self) -> str:
        return os.path.join(self.run_dir, "{top}ILMDir".format(top=self.top_module))

    def write_ilm(self) -> bool:
        """Run time_design and write out the ILM."""
        self.verbose_append("time_design -post_route")
        self.verbose_append("time_design -post_route -hold")
        self.verbose_append("check_process_antenna")
        self.verbose_append("write_lef_abstract -5.8 {top}ILM.lef".format(top=self.top_module))
        self.verbose_append("write_ilm -model_type all -to_dir {ilm_dir_name} -type_flex_ilm ilm".format(
            ilm_dir_name=self.ilm_dir_name))
        self.ran_write_ilm = True
        return True

    def run_innovus(self) -> bool:
        # Quit Innovus.
        self.verbose_append("exit")

        # Create par script.
        par_tcl_filename = os.path.join(self.run_dir, "par.tcl")
        self.write_contents_to_path("\n".join(self.output), par_tcl_filename)

        # Make sure that generated-scripts exists.
        os.makedirs(self.generated_scripts_dir, exist_ok=True)

        # Create open_chip script pointing to latest (symlinked to post_<last ran step>).
        self.output.clear()
        assert super().do_pre_steps(self.first_step)
        self.append("read_db latest")
        self.write_contents_to_path("\n".join(self.output), self.open_chip_tcl)

        with open(self.open_chip_script, "w") as f:
            f.write("""#!/bin/bash
        cd {run_dir}
        source enter
        $INNOVUS_BIN -common_ui -win -files {open_chip_tcl}
                """.format(run_dir=self.run_dir, open_chip_tcl=self.open_chip_tcl))
        os.chmod(self.open_chip_script, 0o755)

        # Build args.
        args = [
            self.get_setting("par.innovus.innovus_bin"),
            "-nowin",  # Prevent the GUI popping up.
            "-common_ui",
            "-files", par_tcl_filename
        ]

        # Temporarily disable colours/tag to make run output more readable.
        # TODO: think of a more elegant way to do this?
        HammerVLSILogging.enable_colour = False
        HammerVLSILogging.enable_tag = False
        self.run_executable(args, cwd=self.run_dir)  # TODO: check for errors and deal with them
        HammerVLSILogging.enable_colour = True
        HammerVLSILogging.enable_tag = True

        # TODO: check that par run was successful

        return True

    def create_floorplan_tcl(self) -> List[str]:
        """
        Create a floorplan TCL depending on the floorplan mode.
        """
        output = []  # type: List[str]

        floorplan_mode = str(self.get_setting("par.innovus.floorplan_mode"))
        if floorplan_mode == "manual":
            floorplan_script_contents = str(self.get_setting("par.innovus.floorplan_script_contents"))
            # TODO(edwardw): proper source locators/SourceInfo
            output.append("# Floorplan manually specified from HAMMER")
            output.extend(floorplan_script_contents.split("\n"))
        elif floorplan_mode == "generate":
            output.extend(self.generate_floorplan_tcl())
        elif floorplan_mode == "auto":
            output.append("# Using auto-generated floorplan")
            output.append("plan_design")
            spacing = self.get_setting("par.blockage_spacing")
            bot_layer = self.get_stackup().get_metal_by_index(1).name
            top_layer = self.get_setting("par.blockage_spacing_top_layer")
            if top_layer is not None:
                output.append("create_place_halo -all_blocks -halo_deltas {{{s} {s} {s} {s}}} -snap_to_site".format(
                    s=spacing))
                output.append("create_route_halo -all_blocks -bottom_layer {b} -space {s} -top_layer {t}".format(
                    b=bot_layer, t=top_layer, s=spacing))
        else:
            if floorplan_mode != "blank":
                self.logger.error("Invalid floorplan_mode {mode}. Using blank floorplan.".format(mode=floorplan_mode))
            # Write blank floorplan
            output.append("# Blank floorplan specified from HAMMER")
        return output

    @staticmethod
    def generate_chip_size_constraint(width: Decimal, height: Decimal, left: Decimal, bottom: Decimal, right: Decimal,
                                      top: Decimal, site: str) -> str:
        """
        Given chip width/height and margins, generate an Innovus TCL command to create the floorplan.
        Also requires a technology specific name for the core site
        """

        site_str = "-site " + site

        # -flip -f allows standard cells to be flipped correctly during place-and-route
        return ("create_floorplan -core_margins_by die -flip f "
                "-die_size_by_io_height max {site_str} "
                "-die_size {{ {width} {height} {left} {bottom} {right} {top} }}").format(
            site_str=site_str,
            width=width,
            height=height,
            left=left,
            bottom=bottom,
            right=right,
            top=top
        )

    def generate_floorplan_tcl(self) -> List[str]:
        """
        Generate a TCL floorplan for Innovus based on the input config/IR.
        Not to be confused with create_floorplan_tcl, which calls this function.
        """
        output = []  # type: List[str]

        # TODO(edwardw): proper source locators/SourceInfo
        output.append("# Floorplan automatically generated from HAMMER")

        # Top-level chip size constraint.
        # Default/fallback constraints if no other constraints are provided.
        # TODO snap this to a core site
        chip_size_constraint = self.generate_chip_size_constraint(
            site=self.technology.get_placement_site().name,
            width=Decimal("1000"), height=Decimal("1000"),
            left=Decimal("100"), bottom=Decimal("100"),
            right=Decimal("100"), top=Decimal("100")
        )

        floorplan_constraints = self.get_placement_constraints()
        global_top_layer = self.get_setting("par.blockage_spacing_top_layer") #  type: Optional[str]

        ############## Actually generate the constraints ################
        for constraint in floorplan_constraints:
            # Floorplan names/insts need to not include the top-level module,
            # despite the internal get_db commands including the top-level module...
            # e.g. Top/foo/bar -> foo/bar
            new_path = "/".join(constraint.path.split("/")[1:])

            if new_path == "":
                assert constraint.type == PlacementConstraintType.TopLevel, "Top must be a top-level/chip size constraint"
                margins = constraint.margins
                assert margins is not None
                # Set top-level chip dimensions.
                chip_size_constraint = self.generate_chip_size_constraint(
                    site=self.technology.get_placement_site().name,
                    width=constraint.width,
                    height=constraint.height,
                    left=margins.left,
                    bottom=margins.bottom,
                    right=margins.right,
                    top=margins.top
                )
            else:
                orientation = constraint.orientation if constraint.orientation is not None else "r0"
                if constraint.create_physical:
                    output.append("create_inst -cell {cell} -inst {inst} -location {{{x} {y}}} -orient {orientation} -physical -status fixed".format(
                        cell=constraint.master,
                        inst=new_path,
                        x=constraint.x,
                        y=constraint.y,
                        orientation=orientation
                    ))
                if constraint.type == PlacementConstraintType.Dummy:
                    pass
                elif constraint.type == PlacementConstraintType.Placement:
                    output.append("create_guide -name {name} -area {x1} {y1} {x2} {y2}".format(
                        name=new_path,
                        x1=constraint.x,
                        x2=constraint.x + constraint.width,
                        y1=constraint.y,
                        y2=constraint.y + constraint.height
                    ))
                elif constraint.type == PlacementConstraintType.Overlap:
                    output.append("place_inst {inst} {x} {y} {orientation}{fixed}".format(
                        inst=new_path,
                        x=constraint.x,
                        y=constraint.y,
                        orientation=orientation,
                        fixed=" -fixed" if constraint.create_physical else ""
                    ))
                elif constraint.type in [PlacementConstraintType.HardMacro, PlacementConstraintType.Hierarchical]:
                    output.append("place_inst {inst} {x} {y} {orientation}{fixed}".format(
                        inst=new_path,
                        x=constraint.x,
                        y=constraint.y,
                        orientation=orientation,
                        fixed=" -fixed" if constraint.create_physical else ""
                    ))
                    spacing = self.get_setting("par.blockage_spacing")
                    if constraint.top_layer is not None:
                        current_top_layer = constraint.top_layer #  type: Optional[str]
                    elif global_top_layer is not None:
                        current_top_layer = global_top_layer
                    else:
                        current_top_layer = None
                    if current_top_layer is not None:
                        bot_layer = self.get_stackup().get_metal_by_index(1).name
                        output.append("create_place_halo -insts {inst} -halo_deltas {{{s} {s} {s} {s}}} -snap_to_site".format(
                            inst=new_path, s=spacing))
                        output.append("create_route_halo -bottom_layer {b} -space {s} -top_layer {t} -inst {inst}".format(
                            inst=new_path, b=bot_layer, t=current_top_layer, s=spacing))
                elif constraint.type == PlacementConstraintType.Obstruction:
                    obs_types = get_or_else(constraint.obs_types, [])  # type: List[ObstructionType]
                    if ObstructionType.Place in obs_types:
                        output.append("create_place_blockage -area {{{x} {y} {urx} {ury}}}".format(
                            x=constraint.x,
                            y=constraint.y,
                            urx=constraint.x+constraint.width,
                            ury=constraint.y+constraint.height
                        ))
                    if ObstructionType.Route in obs_types:
                        output.append("create_route_blockage -layers {{{layers}}} -spacing 0 -{area_flag} {{{x} {y} {urx} {ury}}}".format(
                            x=constraint.x,
                            y=constraint.y,
                            urx=constraint.x+constraint.width,
                            ury=constraint.y+constraint.height,
                            area_flag="rects" if self.version() >= self.version_number("181") else "area",
                            layers="all" if constraint.layers is None else " ".join(get_or_else(constraint.layers, []))
                        ))
                    if ObstructionType.Power in obs_types:
                        output.append("create_route_blockage -pg_nets -layers {{{layers}}} -{area_flag} {{{x} {y} {urx} {ury}}}".format(
                            x=constraint.x,
                            y=constraint.y,
                            urx=constraint.x+constraint.width,
                            ury=constraint.y+constraint.height,
                            area_flag="rects" if self.version() >= self.version_number("181") else "area",
                            layers="all" if constraint.layers is None else " ".join(get_or_else(constraint.layers, []))
                        ))
                else:
                    assert False, "Should not reach here"
        return [chip_size_constraint] + output

    def specify_std_cell_power_straps(self, blockage_spacing: Decimal, bbox: Optional[List[Decimal]], nets: List[str]) -> List[str]:
        """
        Generate a list of TCL commands that build the low-level standard cell power strap rails.
        This will use the -master option to create power straps based on technology.core.tap_cell_rail_reference.
        The layer is set by technology.core.std_cell_rail_layer, which should be the highest metal layer in the std cell rails.

        :param bbox: The optional (2N)-point bounding box of the area to generate straps. By default the entire core area is used.
        :param nets: A list of power net names (e.g. ["VDD", "VSS"]). Currently only two are supported.
        :return: A list of TCL commands that will generate power straps on rails.
        """
        layer_name = self.get_setting("technology.core.std_cell_rail_layer")
        layer = self.get_stackup().get_metal(layer_name)
        results = [
            "# Power strap definition for layer {} (rails):\n".format(layer_name),
            "reset_db -category add_stripes",
            "set_db add_stripes_stacked_via_bottom_layer {}".format(layer_name),
            "set_db add_stripes_stacked_via_top_layer {}".format(layer_name),
            "set_db add_stripes_spacing_from_block {}".format(blockage_spacing)
        ]
        tapcell = self.get_setting("technology.core.tap_cell_rail_reference")
        options = [
            "-pin_layer", layer_name,
            "-layer", layer_name,
            "-over_pins", "1",
            "-master", "\"{}\"".format(tapcell),
            "-block_ring_bottom_layer_limit", layer_name,
            "-block_ring_top_layer_limit", layer_name,
            "-pad_core_ring_bottom_layer_limit", layer_name,
            "-pad_core_ring_top_layer_limit", layer_name,
            "-direction", str(layer.direction),
            "-width", "pin_width",
            "-nets", "{ %s }" % " ".join(nets)
        ]
        if bbox is not None:
            options.extend([
                "-area", "{ %s }" % " ".join(map(str, bbox))
            ])
        results.append("add_stripes " + " ".join(options) + "\n")
        return results

    def specify_power_straps(self, layer_name: str, bottom_via_layer_name: str, blockage_spacing: Decimal, pitch: Decimal, width: Decimal, spacing: Decimal, offset: Decimal, bbox: Optional[List[Decimal]], nets: List[str], add_pins: bool) -> List[str]:
        """
        Generate a list of TCL commands that will create power straps on a given layer.
        This is a low-level, cad-tool-specific API. It is designed to be called by higher-level methods, so calling this directly is not recommended.
        This method assumes that power straps are built bottom-up, starting with standard cell rails.

        :param layer_name: The layer name of the metal on which to create straps.
        :param bottom_via_layer_name: The layer name of the lowest metal layer down to which to drop vias.
        :param blockage_spacing: The minimum spacing between the end of a strap and the beginning of a macro or blockage.
        :param pitch: The pitch between groups of power straps (i.e. from left edge of strap A to the next left edge of strap A).
        :param width: The width of each strap in a group.
        :param spacing: The spacing between straps in a group.
        :param offset: The offset to start the first group.
        :param bbox: The optional (2N)-point bounding box of the area to generate straps. By default the entire core area is used.
        :param nets: A list of power nets to create (e.g. ["VDD", "VSS"], ["VDDA", "VSS", "VDDB"],  ... etc.).
        :param add_pins: True if pins are desired on this layer; False otherwise.
        :return: A list of TCL commands that will generate power straps.
        """
        # TODO check that this has been not been called after a higher-level metal and error if so
        # TODO warn if the straps are off-pitch
        results = ["# Power strap definition for layer %s:\n" % layer_name]
        results.extend([
            "reset_db -category add_stripes",
            "set_db add_stripes_stacked_via_top_layer {}".format(layer_name),
            "set_db add_stripes_stacked_via_bottom_layer {}".format(bottom_via_layer_name),
            "set_db add_stripes_trim_antenna_back_to_shape {stripe}",
            "set_db add_stripes_spacing_from_block {}".format(blockage_spacing)
        ])
        layer = self.get_stackup().get_metal(layer_name)
        options = [
            "-create_pins", ("1" if (add_pins) else "0"),
            "-block_ring_bottom_layer_limit", layer_name,
            "-block_ring_top_layer_limit", bottom_via_layer_name,
            "-direction", str(layer.direction),
            "-layer", layer_name,
            "-nets", "{%s}" % " ".join(nets),
            "-pad_core_ring_bottom_layer_limit", bottom_via_layer_name,
            "-set_to_set_distance", str(pitch),
            "-spacing", str(spacing),
            "-switch_layer_over_obs", "0",
            "-width", str(width)
        ]
        # Where to get the io-to-core offset from a bbox
        index = 0
        if layer.direction == RoutingDirection.Horizontal:
            index = 1
        elif layer.direction != RoutingDirection.Vertical:
            raise ValueError("Cannot handle routing direction {d} for layer {l} when creating power straps".format(d=str(layer.direction), l=layer_name))

        if bbox is not None:
            options.extend([
                "-area", "{ %s }" % " ".join(map(str, bbox)),
                "-start", str(offset + bbox[index])
            ])

        else:
            # Just put straps in the core area
            options.extend([
                "-area", "[get_db designs .core_bbox]",
                "-start", "[expr [lindex [lindex [get_db designs .core_bbox] 0] {index}] + {offset}]".format(index=index, offset=offset)
            ])
        results.append("add_stripes " + " ".join(options) + "\n")
        return results

def innovus_global_settings(ht: HammerTool) -> bool:
    """Settings that need to be reapplied at every tool invocation"""
    assert isinstance(ht, HammerPlaceAndRouteTool)
    assert isinstance(ht, CadenceTool)
    ht.create_enter_script()

    # Python sucks here for verbosity
    verbose_append = ht.verbose_append

    # Generic settings
    verbose_append("set_db design_process_node {}".format(ht.get_setting("vlsi.core.node")))
    verbose_append("set_multi_cpu_usage -local_cpu {}".format(ht.get_setting("vlsi.core.max_threads")))

    return True

tool = Innovus



"""
Convert YAML file to JSON with identical structure (as a python dict)
Adapted from https://raw.githubusercontent.com/vasil9v/yaml2json.py/c4cfa408bbab4f1a0cd3661d121237030f6bc0ca/yaml2json.py
"""
import sys
import yaml
from hammer.config.yaml2json import convertArrays, compare

def load(f):
    try:
        return (open(f, 'r')).read()
    except IOError:
        return None

def save(f, b): (open(f, 'w')).write(b)

def yaml2json():
    """
    Main function, first arg is the input file, optional second one is the output
    file. If no output file is specified then the output is written to stdout.
    The input file is parsed as YAML and converted to a python dict tree, then
    that tree is converted to the JSON output. There is a check to make sure the
    two dict trees are structurally identical.
    """
    if len(sys.argv) > 1:
        f = sys.argv[1]
        f2 = None
        if len(sys.argv) > 2:
            f2 = sys.argv[2]
        obj = yaml.safe_load(load(f))
        obj = convertArrays(obj)
        outputContent = json.dumps(obj, indent=2)
        obj2 = json.loads(outputContent)
        if not compare(obj, obj2):
            print("error: they dont match structure")
            print("")
            print(str(obj))
            print("")
            print(str(obj2))
        else:
            if f2:
                save(f2, outputContent)
            else:
                print(outputContent)
    else:
        print("usage: yaml2json infile.yaml [outfile.json]")

def main():
    yaml2json()
